from __future__ import annotations

import typing as t
from functools import partial

from ..._internal.utils import pkg

if t.TYPE_CHECKING:
    from starlette import requests

if pkg.pkg_version_info("ray")[:2] >= (2, 5):
    from ray.serve._private.http_util import BufferedASGISender as ASGIHTTPSender
else:
    from ray.serve._private.http_util import ASGIHTTPSender

import bentoml
from bentoml import Tag

from ...exceptions import MissingDependencyException
from ..runner.container import AutoContainer
from ..runner.utils import Params
from .runner_handle import RayRunnerHandle

try:
    from ray import serve
    from ray.runtime_env import RuntimeEnv
    from ray.serve.deployment import Deployment
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """'ray[serve]' is required in order to use module 'bentoml.ray', install with 'pip install -U "ray[serve]"'. See https://docs.ray.io/ for more information.""",
    )


def _get_runner_deployment(
    svc: bentoml.Service,
    runner_name: str,
    runner_deployment_config: dict[str, t.Any],
    enable_batching: bool,
    batching_config: dict[str, dict[str, float | int]],
) -> Deployment:
    class RunnerDeployment:
        def __init__(self):
            _runner = next(
                (runner for runner in svc.runners if runner.name == runner_name),
                None,
            )
            _runner.init_local(quiet=True)
            for method in _runner.runner_methods:
                if enable_batching and method.config.batchable:
                    inp_batch_dim = method.config.batch_dim[0]
                    out_batch_dim = method.config.batch_dim[1]
                    ray_batch_args = (
                        batching_config[method.name]
                        if method.name in batching_config
                        else {}
                    )

                    @serve.batch(**ray_batch_args)
                    async def _func(self, *args, **kwargs):
                        params = Params(*args, **kwargs).map(
                            partial(
                                AutoContainer.batches_to_batch, batch_dim=inp_batch_dim
                            )
                        )
                        run_params = params.map(lambda arg: arg[0])
                        indices = next(iter(params.items()))[1][1]
                        results = await getattr(_runner, method.name).async_run(
                            *run_params.args, **run_params.kwargs
                        )
                        return AutoContainer.batch_to_batches(
                            results, indices=indices, batch_dim=out_batch_dim
                        )

                    setattr(RunnerDeployment, method.name, _func)
                else:

                    async def _func(self, *args, **kwargs):
                        return await getattr(_runner, method.name).async_run(
                            *args, **kwargs
                        )

                    setattr(RunnerDeployment, method.name, _func)

    return serve.deployment(
        RunnerDeployment, name=f"bento-runner-{runner_name}", **runner_deployment_config
    )


def _get_service_deployment(svc: bentoml.Service, **kwargs) -> Deployment:
    @serve.deployment(name=f"bento-svc-{svc.name}", **kwargs)
    class BentoDeployment:
        def __init__(self, **runner_deployments: dict[str, Deployment]):
            from ..server.http_app import HTTPAppFactory

            # Use Ray's metrics setup. Ray has already initialized a prometheus client
            # for multi-process and this conflicts with BentoML's prometheus setup
            self.app = HTTPAppFactory(svc, enable_metrics=False)()

            for runner in svc.runners:
                assert runner.name in runner_deployments
                runner._set_handle(RayRunnerHandle, runner_deployments[runner.name])

        async def __call__(self, request: requests.Request):
            sender = ASGIHTTPSender()
            await self.app(request.scope, receive=request.receive, send=sender)
            return sender.build_asgi_response()

    return BentoDeployment


def get_bento_runtime_env(bento_tag: str | Tag) -> RuntimeEnv:
    # TODO: create Ray RuntimeEnv from Bento env configs
    return RuntimeEnv(
        pip=[],
        conda=[],
        env_vars=[],
    )


def _deploy_bento_runners(
    svc: bentoml.Service,
    runners_deployment_config_map: dict | None = None,
    enable_batching: bool = False,
    batching_config: dict | None = None,
) -> dict[str, Deployment]:
    # Deploy BentoML Runners as Ray serve deployments
    runners_deployment_config_map = runners_deployment_config_map or {}
    batching_config = batching_config or {}
    runner_deployments = {}
    for runner in svc.runners:
        runner_deployment_config = (
            runners_deployment_config_map.get(runner.name)
            if runner.name in runners_deployment_config_map
            else {}
        )
        batching_config = (
            batching_config.get(runner.name) if runner.name in batching_config else {}
        )
        runner_deployment = _get_runner_deployment(
            svc, runner.name, runner_deployment_config, enable_batching, batching_config
        ).bind()
        runner_deployments[runner.name] = runner_deployment

    return runner_deployments


def deployment(
    target: str | Tag | bentoml.Bento | bentoml.Service,
    service_deployment_config: dict[str, t.Any] | None = None,
    runners_deployment_config_map: dict[str, dict[str, t.Any]] | None = None,
    enable_batching: bool = False,
    batching_config: dict[str, dict[str, dict[str, float | int]]] | None = None,
) -> Deployment:
    """
    Deploy a Bento or bentoml.Service to Ray

    Args:
        target: A bentoml.Service instance, Bento tag string, or Bento object
        service_deployment_config: Ray deployment config for BentoML API server
        runners_deployment_config_map: Ray deployment config map for all Runners
        enable_batching: Enable Ray Serve batching for RunnerMethods with batchable=True
        batching_config: Pass through Ray batching config by Runner name and method name

    Returns:
        A bound ray.serve.Deployment instance

    Example:

    .. code-block:: python
        :caption: `ray_demo.py`

        import bentoml

        classifier = bentoml.ray.deployment('iris_classifier:latest')

    .. code-block:: bash

        serve run ray_demo:classifier

    Configure BentoML-on-Ray deployment:

    .. code-block:: python

        import bentoml

        classifier = bentoml.ray.deployment(
            'iris_classifier:latest',
            {"route_prefix": "/hello", "num_replicas": 3, "ray_actor_options": {"num_cpus": 1}},
            {"iris_clf": {"num_replicas": 1, "ray_actor_options": {"num_cpus": 5}}}
        )

    Configure BentoML Runner's Batching behavior on Ray:

    .. code-block:: python

        import bentoml

        deploy = bentoml.ray.deployment(
            'fraud_detection:latest',
            {"num_replicas": 5, "ray_actor_options": {"num_cpus": 1}},
            {"ieee-fraud-detection-sm": {"num_replicas": 1, "ray_actor_options": {"num_cpus": 5}}},
            enable_batching=True,
            batching_config={"ieee-fraud-detection-sm": {"predict_proba": {"max_batch_size": 5, "batch_wait_timeout_s": 0.2}}}
        )

    """
    # TODO: validate Ray deployment options
    service_deployment_config = service_deployment_config or {}

    svc = target if isinstance(target, bentoml.Service) else bentoml.load(target)

    runner_deployments = _deploy_bento_runners(
        svc, runners_deployment_config_map, enable_batching, batching_config
    )

    return _get_service_deployment(svc, **service_deployment_config).bind(
        **runner_deployments
    )



from __future__ import annotations

import logging
import os
import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import attr
import numpy as np

import bentoml
from bentoml import Tag
from bentoml._internal.models.model import ModelContext
from bentoml._internal.utils import deprecated
from bentoml._internal.utils.pkg import get_pkg_version
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import InvalidArgument
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import ModelOptions as BaseModelOptions
from bentoml.models import get as get

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from _bentoml_sdk import Service
    from _bentoml_sdk import ServiceConfig
    from bentoml._internal.models.model import ModelSignaturesType
    from bentoml.types import ModelSignature


try:
    import xgboost as xgb
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'xgboost' is required in order to use module 'bentoml.xgboost', install xgboost with 'pip install xgboost'. For more information, refer to https://xgboost.readthedocs.io/en/latest/install.html"
    ) from None

try:
    from xgboost import XGBModel
except ImportError:  # pragma: no cover
    # if sklearn is not installed, XGBoost will not expose XGBModel, so make
    # a dummy class ourselves
    class XGBModel:
        pass


MODULE_NAME = "bentoml.xgboost"
MODEL_FILENAME = "saved_model.ubj"
API_VERSION = "v2"

logger = logging.getLogger(__name__)


@attr.define
class ModelOptions(BaseModelOptions):
    model_class: str | None = None


def load_model(bento_model: str | Tag | bentoml.Model) -> xgb.Booster | xgb.XGBModel:
    """
    Load the XGBoost model with the given tag from the local BentoML model store.

    Args:
        bento_model (``str`` ``|`` :obj:`~bentoml.Tag` ``|`` :obj:`~bentoml.Model`):
            Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
            instance to load the model from.
    Returns:
        The XGBoost model loaded from the model store or BentoML :obj:`~bentoml.Model`.
    Example:

    .. code-block:: python

        import bentoml
        # target model must be from the BentoML model store
        booster = bentoml.xgboost.load_model("my_xgboost_model")
    """  # noqa: LN001
    if not isinstance(bento_model, bentoml.Model):
        bento_model = bentoml.models.get(bento_model)
        assert isinstance(bento_model, bentoml.Model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    model_file = bento_model.path_of(MODEL_FILENAME)
    model_api_version = bento_model.info.api_version
    if model_api_version == "v1":
        model = xgb.Booster(model_file=model_file)
    else:
        if model_api_version != "v2":
            logger.warning(
                "Got an XGBoost model with an unsupported version '%s', unexpected errors may occur.",
                model_api_version,
            )
        model_class = t.cast(ModelOptions, bento_model.info.options).model_class
        if model_class is None:
            raise BentoMLException(
                f"Model '{bento_model.tag}' is missing the required 'model_class' option. This should not be possible; please file an issue if you encounter this error."
            )
        try:
            xgb_class: type[xgb.XGBModel] | type[xgb.Booster] = getattr(
                xgb, model_class
            )
        except AttributeError:
            if model_class != "Booster":
                raise BentoMLException(
                    f"Model '{bento_model.tag}' is an XGBoost Scikit-Learn model, but sklearn is not installed."
                ) from None
            else:
                raise BentoMLException(
                    "xgboost.Booster could not be found, your XGBoost installation may be corrupted. Ensure there is no file named 'xgboost.py' that may be being loaded instead of the XGBoost library."
                ) from None
        model: xgb.Booster | xgb.XGBModel = xgb_class()
        model.load_model(model_file)
    return model


def save_model(
    name: Tag | str,
    model: xgb.Booster | xgb.XGBModel,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save an XGBoost model instance to the BentoML model store.

    Args:
        name:
            The name to give to the model in the BentoML store. This must be a valid
            :obj:`~bentoml.Tag` name.
        model:
            The XGBoost model to be saved.
        signatures:
            Signatures of predict methods to be used. If not provided, the signatures default to
            ``{"predict": {"batchable": False}}``. See :obj:`~bentoml.types.ModelSignature` for more
            details.
        labels:
            A default set of management labels to be associated with the model. An example is
            ``{"training-set": "data-1"}``.
        custom_objects:
            Custom objects to be saved with the model. An example is
            ``{"my-normalizer": normalizer}``.

            Custom objects are currently serialized with cloudpickle, but this implementation is
            subject to change.
        external_modules:
            user-defined additional python modules to be saved alongside the model or custom objects,
            e.g. a tokenizer module, preprocessor module, model configuration module
        metadata:
            Metadata to be associated with the model. An example is ``{"max_depth": 2}``.

            Metadata is intended for display in model management UI and therefore must be a default
            Python type, such as ``str`` or ``int``.
    Returns:
        A BentoML tag with the user-defined name and a generated version.

    Example:

    .. code-block:: python

        import xgboost as xgb
        import bentoml

        # read in data
        dtrain = xgb.DMatrix('demo/data/agaricus.txt.train')
        dtest = xgb.DMatrix('demo/data/agaricus.txt.test')
        # specify parameters via map
        param = dict(max_depth=2, eta=1, objective='binary:logistic')
        num_round = 2
        bst = xgb.train(param, dtrain, num_round)
        ...

        # `save` the booster to BentoML modelstore:
        bento_model = bentoml.xgboost.save_model("my_xgboost_model", bst, booster_params=param)
    """  # noqa: LN001
    if isinstance(model, xgb.Booster):
        model_class = "Booster"
    elif isinstance(model, XGBModel):
        model_class = model.__class__.__name__
    else:
        raise TypeError(f"Given model ({model}) is not a xgboost.Booster.")

    context = ModelContext(
        framework_name="xgboost",
        framework_versions={"xgboost": get_pkg_version("xgboost")},
    )
    if signatures is None:
        signatures = {"predict": {"batchable": False}}
        logger.info(
            'Using the default model signature for xgboost (%s) for model "%s".',
            signatures,
            name,
        )

    with bentoml.models._create(  # type: ignore
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        signatures=signatures,
        labels=labels,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        context=context,
        options=ModelOptions(model_class=model_class),
    ) as bento_model:
        model.save_model(bento_model.path_of(MODEL_FILENAME))  # type: ignore (incomplete XGBoost types)

        return bento_model


@deprecated(suggestion="Use `get_service` instead.")
def get_runnable(bento_model: bentoml.Model) -> t.Type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class XGBoostRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

            self.booster = (
                self.model
                if isinstance(self.model, xgb.Booster)
                else self.model.get_booster()
            )

            # check for resources
            if os.getenv("CUDA_VISIBLE_DEVICES") not in (None, "", "-1"):
                self.booster.set_param({"predictor": "gpu_predictor", "gpu_id": 0})  # type: ignore (incomplete XGBoost types)
            else:
                nthreads = os.getenv("OMP_NUM_THREADS")
                if nthreads is not None and nthreads != "":
                    nthreads = max(int(nthreads), 1)
                else:
                    nthreads = 1
                self.booster.set_param(
                    {"predictor": "cpu_predictor", "nthread": nthreads}
                )  # type: ignore (incomplete XGBoost types)

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                try:
                    self.predict_fns[method_name] = getattr(self.model, method_name)
                except AttributeError:
                    raise InvalidArgument(
                        f"No method with name {method_name} found for XGBoost model of type {self.model.__class__}"
                    )

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(
            self: XGBoostRunnable, input_data: t.Any, *args: t.Any, **kwargs: t.Any
        ) -> t.Any:
            if isinstance(self.model, xgb.Booster):
                inp = xgb.DMatrix(input_data)
            else:
                inp = input_data

            res = self.predict_fns[method_name](inp, *args, **kwargs)
            return np.asarray(res)  # type: ignore (incomplete np types)

        XGBoostRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return XGBoostRunnable


def get_service(model_name: str, **config: Unpack[ServiceConfig]) -> Service[t.Any]:
    """
    Get a BentoML service instance from an XGBoost model.

    Args:
        model_name (``str``):
            The name of the model to get the service for.
        **config (``Unpack[ServiceConfig]``):
            Configuration options for the service.
    Returns:
        A BentoML service instance that wraps the XGBoost model.
    Example:

    .. code-block:: python

        import bentoml

        service = bentoml.xgboost.get_service("my_xgboost_model")
    """

    @bentoml.service(**config)
    class XGBoostService:
        bento_model = bentoml.models.get(model_name)

        def __init__(self) -> None:
            self.model = load_model(self.bento_model)
            self.booster = (
                self.model
                if isinstance(self.model, xgb.Booster)
                else self.model.get_booster()
            )

            # check for resources
            if os.getenv("CUDA_VISIBLE_DEVICES") not in (None, "", "-1"):
                self.booster.set_param({"predictor": "gpu_predictor", "gpu_id": 0})  # type: ignore (incomplete XGBoost types)
            else:
                nthreads = os.getenv("OMP_NUM_THREADS")
                if nthreads is not None and nthreads != "":
                    nthreads = max(int(nthreads), 1)
                else:
                    nthreads = 1
                self.booster.set_param(
                    {"predictor": "cpu_predictor", "nthread": nthreads}
                )  # type: ignore (incomplete XGBoost types)

        if bento_model.info.options.model_class == "Booster":

            @bentoml.api
            def predict(
                self,
                data: np.ndarray,
                output_margin: bool = False,
                pred_leaf: bool = False,
                pred_contribs: bool = False,
                approx_contribs: bool = False,
                pred_interactions: bool = False,
                validate_features: bool = True,
                training: bool = False,
                iteration_range: t.Tuple[int, int] = (0, 0),
                strict_shape: bool = False,
            ) -> np.ndarray:
                assert isinstance(self.model, xgb.Booster)
                return self.model.predict(
                    xgb.DMatrix(data),
                    output_margin=output_margin,
                    pred_leaf=pred_leaf,
                    pred_contribs=pred_contribs,
                    approx_contribs=approx_contribs,
                    pred_interactions=pred_interactions,
                    validate_features=validate_features,
                    training=training,
                    iteration_range=iteration_range,
                    strict_shape=strict_shape,
                )  # type: ignore (incomplete XGBoost types)
        else:

            @bentoml.api
            def predict(
                self,
                data: np.ndarray,
                output_margin: bool = False,
                validate_features: bool = True,
                base_margin: t.Optional[t.List[t.Any]] = None,
                iteration_range: t.Optional[t.Tuple[int, int]] = None,
            ) -> np.ndarray:
                assert isinstance(self.model, XGBModel)
                return self.model.predict(
                    data,
                    output_margin=output_margin,
                    validate_features=validate_features,
                    base_margin=base_margin,
                    iteration_range=iteration_range,
                )

    return XGBoostService



from __future__ import annotations

import click
import psutil


def create_bentoml_cli() -> click.Command:
    from bentoml._internal.configuration import BENTOML_VERSION
    from bentoml._internal.context import server_context
    from bentoml_cli.bentos import bento_command
    from bentoml_cli.cloud import cloud_command
    from bentoml_cli.containerize import containerize_command
    from bentoml_cli.deployment import deploy_command
    from bentoml_cli.deployment import deployment_command
    from bentoml_cli.deployment import develop_command
    from bentoml_cli.env import env_command
    from bentoml_cli.models import model_command
    from bentoml_cli.secret import secret_command
    from bentoml_cli.serve import serve_command
    from bentoml_cli.start import start_command
    from bentoml_cli.utils import BentoMLCommandGroup
    from bentoml_cli.utils import get_entry_points

    server_context.service_type = "cli"

    CONTEXT_SETTINGS = {"help_option_names": ("-h", "--help")}

    @click.group(cls=BentoMLCommandGroup, context_settings=CONTEXT_SETTINGS)
    @click.version_option(BENTOML_VERSION, "-v", "--version")
    def bentoml_cli():
        """
        \b
        ██████╗ ███████╗███╗   ██╗████████╗ ██████╗ ███╗   ███╗██╗
        ██╔══██╗██╔════╝████╗  ██║╚══██╔══╝██╔═══██╗████╗ ████║██║
        ██████╔╝█████╗  ██╔██╗ ██║   ██║   ██║   ██║██╔████╔██║██║
        ██╔══██╗██╔══╝  ██║╚██╗██║   ██║   ██║   ██║██║╚██╔╝██║██║
        ██████╔╝███████╗██║ ╚████║   ██║   ╚██████╔╝██║ ╚═╝ ██║███████╗
        ╚═════╝ ╚══════╝╚═╝  ╚═══╝   ╚═╝    ╚═════╝ ╚═╝     ╚═╝╚══════╝
        """

    # Add top-level CLI commands
    bentoml_cli.add_command(env_command)
    bentoml_cli.add_command(cloud_command)
    bentoml_cli.add_command(model_command)
    bentoml_cli.add_subcommands(bento_command)
    bentoml_cli.add_subcommands(start_command)
    bentoml_cli.add_subcommands(serve_command)
    bentoml_cli.add_command(containerize_command)
    bentoml_cli.add_command(deploy_command)
    bentoml_cli.add_command(develop_command)
    bentoml_cli.add_command(deployment_command)
    bentoml_cli.add_command(secret_command)
    # Load commands from extensions
    for ep in get_entry_points("bentoml.commands"):
        bentoml_cli.add_command(ep.load())

    if psutil.WINDOWS:
        import sys

        sys.stdout.reconfigure(encoding="utf-8")  # type: ignore

    return bentoml_cli


cli = create_bentoml_cli()


if __name__ == "__main__":
    cli()



from __future__ import annotations

import importlib.metadata
import importlib.util
import logging
import re
import typing as t
from contextlib import contextmanager
from pathlib import Path

import torch
import torch.nn as nn
from packaging import version

logger = logging.getLogger(__name__)


def is_flax_available():
    _flax_available = (
        importlib.util.find_spec("jax") is not None
        and importlib.util.find_spec("flax") is not None
    )
    if _flax_available:
        try:
            importlib.metadata.version("jax")
            importlib.metadata.version("flax")
        except importlib.metadata.PackageNotFoundError:
            _flax_available = False
    return _flax_available


def is_torch_available():
    _torch_available = importlib.util.find_spec("torch") is not None
    if _torch_available:
        try:
            importlib.metadata.version("torch")
        except importlib.metadata.PackageNotFoundError:
            _torch_available = False
    return _torch_available


def is_tf_available():
    from .tensorflow import get_tf_version

    _tf_available = importlib.util.find_spec("tensorflow") is not None
    if _tf_available:
        _tf_version = get_tf_version()
        _tf_available = _tf_version != ""
        if version.parse(_tf_version) < version.parse("2"):
            logger.info(
                f"Tensorflow found but with verison {_tf_version}. Transformers support with BentoML requires a minimum of Tensorflow 2 and above."
            )
            _tf_available = False
    return _tf_available


def extract_commit_hash(
    resolved_dir: str, regex_commit_hash: t.Pattern[str]
) -> str | None:
    """
    Extracts the commit hash from a resolved filename toward a cache file.
    modified from https://github.com/huggingface/transformers/blob/0b7b4429c78de68acaf72224eb6dae43616d820c/src/transformers/utils/hub.py#L219
    """

    resolved_dir = str(Path(resolved_dir).as_posix()) + "/"
    search = re.search(r"snapshots/([^/]+)/", resolved_dir)

    if search is None:
        return None

    commit_hash = search.groups()[0]
    return commit_hash if regex_commit_hash.match(commit_hash) else None


@contextmanager
def init_empty_weights(include_buffers: bool = False):
    """
    init_empty_weights is vendored from accelerate. This is a useful function as it allows us to call transformers.AutoModel.from_pretrained without loading
    the actual weights of the model. This grants access to the model class and its attributes without using a significant amount of memory.

    A context manager under which models are initialized with all parameters on the meta device, therefore creating an
    empty model. Useful when just initializing the model would blow the available RAM.

    Args:
        include_buffers (`bool`, *optional*, defaults to `False`):
            Whether or not to also put all buffers on the meta device while initializing.

    Example:

    ```pyton
    import torch.nn as nn
    from accelerate import init_empty_weights

    # Initialize a model with 100 billions parameters in no time and without using any RAM.
    with init_empty_weights():
        tst = nn.Sequential(*[nn.Linear(10000, 10000) for _ in range(1000)])
    ```

    <Tip warning={true}>

    Any model created under this context manager has no weights. As such you can't do something like
    `model.to(some_device)` with it. To load weights inside your empty model, see [`load_checkpoint_and_dispatch`].

    </Tip>
    """
    old_register_parameter = nn.Module.register_parameter
    if include_buffers:
        old_register_buffer = nn.Module.register_buffer

    def register_empty_parameter(module, name, param):
        old_register_parameter(module, name, param)
        if param is not None:
            param_cls = type(module._parameters[name])
            kwargs = module._parameters[name].__dict__
            module._parameters[name] = param_cls(
                module._parameters[name].to(torch.device("meta")), **kwargs
            )

    def register_empty_buffer(module, name, buffer):
        old_register_buffer(module, name, buffer)
        if buffer is not None:
            module._buffers[name] = module._buffers[name].to(torch.device("meta"))

    try:
        nn.Module.register_parameter = register_empty_parameter
        if include_buffers:
            nn.Module.register_buffer = register_empty_buffer
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter
        if include_buffers:
            nn.Module.register_buffer = old_register_buffer



from __future__ import annotations

import json
import os
import subprocess
import typing as t

import click
import click_option_group as cog
import rich
import yaml
from rich.syntax import Syntax
from rich.table import Table
from simple_di import Provide
from simple_di import inject

from bentoml_cli.utils import is_valid_bento_name
from bentoml_cli.utils import is_valid_bento_tag

if t.TYPE_CHECKING:
    from click import Context
    from click import Parameter

    from bentoml._internal.bento import BentoStore
    from bentoml._internal.cloud import BentoCloudClient
    from bentoml._internal.container import DefaultBuilder


BENTOML_FIGLET = """
██████╗ ███████╗███╗   ██╗████████╗ ██████╗ ███╗   ███╗██╗
██╔══██╗██╔════╝████╗  ██║╚══██╔══╝██╔═══██╗████╗ ████║██║
██████╔╝█████╗  ██╔██╗ ██║   ██║   ██║   ██║██╔████╔██║██║
██╔══██╗██╔══╝  ██║╚██╗██║   ██║   ██║   ██║██║╚██╔╝██║██║
██████╔╝███████╗██║ ╚████║   ██║   ╚██████╔╝██║ ╚═╝ ██║███████╗
╚═════╝ ╚══════╝╚═╝  ╚═══╝   ╚═╝    ╚═════╝ ╚═╝     ╚═╝╚══════╝
"""

ALLOWED_PLATFORMS = [
    "windows",
    "linux",
    "macos",
    "x86_64-pc-windows-msvc",
    "i686-pc-windows-msvc",
    "x86_64-unknown-linux-gnu",
    "aarch64-apple-darwin",
    "x86_64-apple-darwin",
    "aarch64-unknown-linux-gnu",
    "aarch64-unknown-linux-musl",
    "x86_64-unknown-linux-musl",
    "x86_64-manylinux_2_17",
    "x86_64-manylinux_2_28",
    "x86_64-manylinux_2_31",
    "x86_64-manylinux_2_32",
    "x86_64-manylinux_2_33",
    "x86_64-manylinux_2_34",
    "x86_64-manylinux_2_35",
    "x86_64-manylinux_2_36",
    "x86_64-manylinux_2_37",
    "x86_64-manylinux_2_38",
    "x86_64-manylinux_2_39",
    "x86_64-manylinux_2_40",
    "aarch64-manylinux_2_17",
    "aarch64-manylinux_2_28",
    "aarch64-manylinux_2_31",
    "aarch64-manylinux_2_32",
    "aarch64-manylinux_2_33",
    "aarch64-manylinux_2_34",
    "aarch64-manylinux_2_35",
    "aarch64-manylinux_2_36",
    "aarch64-manylinux_2_37",
    "aarch64-manylinux_2_38",
    "aarch64-manylinux_2_39",
    "aarch64-manylinux_2_40",
]


def parse_delete_targets_argument_callback(
    ctx: Context,
    params: Parameter,
    value: t.Any,  # pylint: disable=unused-argument
) -> list[str]:
    if value is None:
        return value
    value = " ".join(value)
    if "," in value:
        delete_targets = value.split(",")
    else:
        delete_targets = value.split()
    delete_targets = list(map(str.strip, delete_targets))
    for delete_target in delete_targets:
        if not (
            is_valid_bento_tag(delete_target) or is_valid_bento_name(delete_target)
        ):
            raise click.BadParameter(
                f'Bad formatting: "{delete_target}". Please present a valid bento bundle name or "name:version" tag. For list of bento bundles, separate delete targets by ",", for example: "my_service:v1,my_service:v2,classifier"'
            )
    return delete_targets


def bento_management_commands() -> click.Group:
    import bentoml
    from bentoml import Tag
    from bentoml._internal.configuration import get_quiet_mode
    from bentoml._internal.configuration.containers import BentoMLContainer
    from bentoml._internal.utils import human_readable_size
    from bentoml.bentos import build_bentofile
    from bentoml.bentos import import_bento
    from bentoml_cli.utils import BentoMLCommandGroup

    @click.group(cls=BentoMLCommandGroup)
    def bentos():
        """Commands for managing Bento bundles."""
        pass

    @bentos.command()
    @click.argument("bento_tag", type=click.STRING)
    @click.option(
        "-o",
        "--output",
        type=click.Choice(["json", "yaml", "path"]),
        default="yaml",
    )
    @inject
    def get(
        bento_tag: str,
        output: str,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> None:  # type: ignore (not accessed)
        """Print Bento details by providing the bento_tag.

        \b
        bentoml get iris_classifier:qojf5xauugwqtgxi
        bentoml get iris_classifier:qojf5xauugwqtgxi --output=json
        """
        bento = bento_store.get(bento_tag)

        if output == "path":
            rich.print(bento.path)
        elif output == "json":
            info = json.dumps(bento.info.to_dict(), indent=2, default=str)
            rich.print_json(info)
        else:
            info = yaml.dump(bento.info.to_dict(), indent=2, sort_keys=False)
            rich.print(Syntax(info, "yaml", background_color="default"))

    @bentos.command(name="list")
    @click.argument("bento_name", type=click.STRING, required=False)
    @click.option(
        "-o",
        "--output",
        type=click.Choice(["json", "yaml", "table"]),
        default="table",
    )
    @inject
    def list_bentos(
        bento_name: str,
        output: str,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> None:  # type: ignore (not accessed)
        """List Bentos in local store

        \b
        # show all bentos saved
        $ bentoml list

        \b
        # show all versions of bento with the name FraudDetector
        $ bentoml list FraudDetector
        """
        bentos = bento_store.list(bento_name)
        res: list[dict[str, str]] = []
        for bento in sorted(bentos, key=lambda x: x.info.creation_time, reverse=True):
            bento_size = bento.file_size
            model_size = bento.total_size() - bento_size
            res.append(
                {
                    "tag": str(bento.tag),
                    "size": human_readable_size(bento_size),
                    "model_size": human_readable_size(model_size),
                    "creation_time": bento.info.creation_time.astimezone().strftime(
                        "%Y-%m-%d %H:%M:%S"
                    ),
                }
            )

        if output == "json":
            info = json.dumps(res, indent=2)
            rich.print(info)
        elif output == "yaml":
            info = t.cast(str, yaml.safe_dump(res, indent=2))
            rich.print(Syntax(info, "yaml", background_color="default"))
        else:
            table = Table(box=None)
            table.add_column("Tag")
            table.add_column("Size")
            table.add_column("Model Size")
            table.add_column("Creation Time")
            for bento in res:
                table.add_row(
                    bento["tag"],
                    bento["size"],
                    bento["model_size"],
                    bento["creation_time"],
                )
            rich.print(table)

    @bentos.command()
    @click.argument(
        "delete_targets",
        nargs=-1,
        callback=parse_delete_targets_argument_callback,
        required=True,
    )
    @click.option(
        "-y",
        "--yes",
        "--assume-yes",
        is_flag=True,
        help="Skip confirmation when deleting a specific bento bundle",
    )
    @inject
    def delete(
        delete_targets: list[str],
        yes: bool,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> None:  # type: ignore (not accessed)
        """Delete Bento in local bento store.

        \b
        Examples:
            * Delete single bento bundle by "name:version", e.g: `bentoml delete IrisClassifier:v1`
            * Bulk delete all bento bundles with a specific name, e.g.: `bentoml delete IrisClassifier`
            * Bulk delete multiple bento bundles by name and version, separated by ",", e.g.: `bentoml delete Irisclassifier:v1,MyPredictService:v2`
            * Bulk delete multiple bento bundles by name and version, separated by " ", e.g.: `bentoml delete Irisclassifier:v1 MyPredictService:v2`
            * Bulk delete without confirmation, e.g.: `bentoml delete IrisClassifier --yes`
        """

        def delete_target(target: str) -> None:
            tag = Tag.from_str(target)

            if tag.version is None:
                to_delete_bentos = bento_store.list(target)
            else:
                to_delete_bentos = [bento_store.get(tag)]

            for bento in to_delete_bentos:
                if yes:
                    delete_confirmed = True
                else:
                    delete_confirmed = click.confirm(f"delete bento {bento.tag}?")

                if delete_confirmed:
                    bento_store.delete(bento.tag)
                    rich.print(f"{bento} deleted.")

        for target in delete_targets:
            delete_target(target)

    @bentos.command()
    @click.argument("bento_tag", type=click.STRING)
    @click.argument(
        "out_path",
        type=click.STRING,
        default="",
        required=False,
    )
    @inject
    def export(
        bento_tag: str,
        out_path: str,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> None:  # type: ignore (not accessed)
        """Export a Bento to an external file archive

        \b
        Arguments:
            BENTO_TAG: bento identifier
            OUT_PATH: output path of exported bento.

        If out_path argument is not provided, bento is exported to name-version.bento in the current directory.
        Beside the native .bento format, we also support ('tar'), tar.gz ('gz'), tar.xz ('xz'), tar.bz2 ('bz2'), and zip.

        \b
        Examples:
            bentoml export FraudDetector:20210709_DE14C9
            bentoml export FraudDetector:20210709_DE14C9 ./my_bento.bento
            bentoml export FraudDetector:latest ./my_bento.bento
            bentoml export FraudDetector:latest s3://mybucket/bentos/my_bento.bento
        """
        bento = bento_store.get(bento_tag)
        out_path = bento.export(out_path)
        rich.print(f"{bento} exported to {out_path}.")

    @bentos.command(name="import")
    @click.argument("bento_path", type=click.STRING)
    def import_bento_(bento_path: str) -> None:  # type: ignore (not accessed)
        """Import a previously exported Bento archive file

        \b
        Arguments:
            BENTO_PATH: path of Bento archive file

        \b
        Examples:
            bentoml import ./my_bento.bento
            bentoml import s3://mybucket/bentos/my_bento.bento
        """
        bento = import_bento(bento_path)
        rich.print(f"{bento} imported.")

    @bentos.command()
    @click.argument("bento_tag", type=click.STRING)
    @click.option(
        "-f",
        "--force",
        is_flag=True,
        default=False,
        help="Force pull from remote Bento store to local and overwrite even if it already exists in local",
    )
    @click.option("--with-models", is_flag=True, default=False, help="Pull models too")
    @inject
    def pull(
        bento_tag: str,
        force: bool,
        with_models: bool,
        cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
    ) -> None:  # type: ignore (not accessed)
        """Pull Bento from a remote Bento store server."""
        cloud_client.bento.pull(bento_tag, force=force, with_models=with_models)

    @bentos.command()
    @click.argument("bento_tag", type=click.STRING)
    @click.option(
        "-f",
        "--force",
        is_flag=True,
        default=False,
        help="Forced push to remote Bento store even if it exists in remote",
    )
    @click.option(
        "-t",
        "--threads",
        default=10,
        help="Number of threads to use for upload",
    )
    @inject
    def push(
        bento_tag: str,
        force: bool,
        threads: int,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
        cloud_client: BentoCloudClient = Provide[BentoMLContainer.bentocloud_client],
    ) -> None:  # type: ignore (not accessed)
        """Push Bento to a remote Bento store server."""
        bento_obj = bento_store.get(bento_tag)
        if not bento_obj:
            raise click.ClickException(f"Bento {bento_tag} not found in local store")
        cloud_client.bento.push(bento_obj, force=force, threads=threads)

    @bentos.command()
    @click.argument("build_ctx", type=click.Path(), default=".")
    @click.option(
        "-f", "--bentofile", help="Path to bentofile. Default to 'bentofile.yaml'"
    )
    @click.option(
        "--version",
        type=click.STRING,
        default=None,
        help="Bento version. By default the version will be generated.",
    )
    @click.option(
        "--label",
        "labels",
        type=click.STRING,
        multiple=True,
        help="(multiple)Bento labels",
        metavar="KEY=VALUE",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Choice(["tag", "default"]),
        default="default",
        show_default=True,
        help="Output log format. '-o tag' to display only bento tag.",
    )
    @cog.optgroup.group(cls=cog.MutuallyExclusiveOptionGroup, name="Utilities options")
    @cog.optgroup.option(
        "--containerize",
        default=False,
        is_flag=True,
        type=click.BOOL,
        help="Whether to containerize the Bento after building. '--containerize' is the shortcut of 'bentoml build && bentoml containerize'.",
    )
    @cog.optgroup.option(
        "--push",
        default=False,
        is_flag=True,
        type=click.BOOL,
        help="Whether to push the result bento to BentoCloud. Make sure to login with 'bentoml cloud login' first.",
    )
    @click.option(
        "--force", is_flag=True, default=False, help="Forced push to BentoCloud"
    )
    @click.option("--threads", default=10, help="Number of threads to use for upload")
    @click.option(
        "--platform",
        default=None,
        help="Platform to build for",
        type=click.Choice(ALLOWED_PLATFORMS),
    )
    def build(  # type: ignore (not accessed)
        build_ctx: str,
        bentofile: str | None,
        version: str | None,
        labels: tuple[str, ...],
        output: t.Literal["tag", "default"],
        push: bool,
        force: bool,
        threads: int,
        containerize: bool,
        platform: str | None,
    ):
        """Build a new Bento from current directory."""
        from bentoml._internal.configuration import set_quiet_mode
        from bentoml._internal.log import configure_logging

        if output == "tag":
            set_quiet_mode()
            configure_logging()

        labels_dict: dict[str, t.Any] = {}
        for label in labels:
            key, label_value = label.split("=", 1)
            labels_dict[key] = label_value

        service: str | None = None
        if ":" in build_ctx:
            service = build_ctx
            build_ctx = "."

        bento = build_bentofile(
            bentofile,
            service=service,
            version=version,
            labels=labels_dict or None,
            build_ctx=build_ctx,
            platform=platform,
        )

        containerize_cmd = f"bentoml containerize {bento.tag}"
        push_cmd = f"bentoml push {bento.tag}"

        # NOTE: Don't remove the return statement here, since we will need this
        # for usage stats collection if users are opt-in.
        if output == "tag":
            rich.print(f"__tag__:{bento.tag}")
        else:
            if not get_quiet_mode():
                rich.print(BENTOML_FIGLET)
                rich.print(f"[green]Successfully built {bento}.")
                next_steps = [
                    f"\n\n* Deploy to BentoCloud:\n    $ bentoml deploy {bento.tag} -n ${{DEPLOYMENT_NAME}}",
                    "\n\n* Update an existing deployment on BentoCloud:\n"
                    f"    $ bentoml deployment update --bento {bento.tag} ${{DEPLOYMENT_NAME}}",
                ]
                if not containerize:
                    next_steps.append(
                        "\n\n* Containerize your Bento with `bentoml containerize`:\n"
                        f"    $ {containerize_cmd} [or bentoml build --containerize]"
                    )
                if not push:
                    next_steps.append(
                        "\n\n* Push to BentoCloud with `bentoml push`:\n"
                        f"    $ {push_cmd} [or bentoml build --push]"
                    )

                if next_steps:
                    rich.print(f"\n[blue]Next steps: {''.join(next_steps)}[/]")
        if push:
            if not get_quiet_mode():
                rich.print(f"\n[magenta]Pushing {bento} to BentoCloud...[/]")
            cloud_client = BentoMLContainer.bentocloud_client.get()
            cloud_client.bento.push(bento, force=force, threads=threads)
        elif containerize:
            backend: DefaultBuilder = t.cast(
                "DefaultBuilder", os.getenv("BENTOML_CONTAINERIZE_BACKEND", "docker")
            )
            try:
                bentoml.container.health(backend)
            except subprocess.CalledProcessError:
                raise bentoml.exceptions.BentoMLException(
                    f"Backend {backend} is not healthy"
                )
            bentoml.container.build(bento.tag, backend=backend)

        return bento

    return bentos


bento_command = bento_management_commands()



from __future__ import annotations

import logging
import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import bentoml
from bentoml import Tag
from bentoml._internal.types import LazyType
from bentoml._internal.utils import deprecated
from bentoml._internal.utils.pkg import get_pkg_version
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import Model
from bentoml.models import ModelContext
from bentoml.models import get as get

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator
    from sklearn.pipeline import Pipeline

    from bentoml._internal.models.model import ModelSignaturesType
    from bentoml.types import ModelSignature

    SklearnModel: t.TypeAlias = BaseEstimator | Pipeline


try:
    import joblib
    from joblib import parallel_backend
except ImportError:  # pragma: no cover
    try:
        from sklearn.utils._joblib import joblib
        from sklearn.utils._joblib import parallel_backend
    except ImportError:
        raise MissingDependencyException(
            "scikit-learn is required in order to use the module 'bentoml.sklearn', install scikit-learn with 'pip install scikit-learn'. For more information, refer to https://scikit-learn.org/stable/install.html"
        )

MODULE_NAME = "bentoml.sklearn"
MODEL_FILENAME = "saved_model.pkl"
API_VERSION = "v1"

logger = logging.getLogger(MODULE_NAME)


def load_model(bento_model: str | Tag | Model) -> SklearnModel:
    """
    Load the scikit-learn model with the given tag from the local BentoML model store.

    Args:
        bento_model: Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
                     instance to load the model from.

    Returns:
        The scikit-learn model loaded from the model store or BentoML :obj:`~bentoml.Model`.

    Example:

    .. code-block:: python

        import bentoml
        sklearn = bentoml.sklearn.load_model('my_model:latest')
    """  # noqa
    if not isinstance(bento_model, Model):
        bento_model = bentoml.models.get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )
    model_file = bento_model.path_of(MODEL_FILENAME)

    return joblib.load(model_file)


def save_model(
    name: Tag | str,
    model: SklearnModel,
    *,
    signatures: ModelSignaturesType | None = None,
    labels: t.Dict[str, str] | None = None,
    custom_objects: t.Dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: t.Dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name: Name for given model instance. This should pass Python identifier check.
        model: Instance of model to be saved.
        signatures: Methods to expose for running inference on the target model. Signatures are
                    used for creating Runner instances when serving model with bentoml.Service
        labels: user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects: user-defined additional python objects to be saved alongside the model,
                        e.g. a tokenizer instance, preprocessor function, model configuration json
        external_modules: user-defined additional python modules to be saved alongside the model or custom objects,
                          e.g. a tokenizer module, preprocessor module, model configuration module
        metadata: Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format ``name:version`` where ``name`` is
        the user-defined model's name, and a generated ``version``.

    Examples:

    .. code-block:: python

        import bentoml

        from sklearn.datasets import load_iris
        from sklearn.neighbors import KNeighborsClassifier

        model = KNeighborsClassifier()
        iris = load_iris()
        X = iris.data[:, :4]
        Y = iris.target
        model.fit(X, Y)

        bento_model = bentoml.sklearn.save_model('kneighbors', model)
    """  # noqa
    if not (
        LazyType("sklearn.base.BaseEstimator").isinstance(model)
        or LazyType("sklearn.pipeline.Pipeline").isinstance(model)
    ):
        raise TypeError(
            f"Given model ({model}) is not a sklearn.base.BaseEstimator or sklearn.pipeline.Pipeline."
        )

    context = ModelContext(
        framework_name="sklearn",
        framework_versions={"scikit-learn": get_pkg_version("scikit-learn")},
    )

    if signatures is None:
        signatures = {"predict": {"batchable": False}}
        logger.info(
            'Using the default model signature for scikit-learn (%s) for model "%s".',
            signatures,
            name,
        )

    with bentoml.models._create(  # type: ignore
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        labels=labels,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        context=context,
        signatures=signatures,
    ) as bento_model:
        joblib.dump(model, bento_model.path_of(MODEL_FILENAME))

        return bento_model


@deprecated(suggestion="Use `get_service` instead.")
def get_runnable(bento_model: Model):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class SklearnRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("cpu",)
        SUPPORTS_CPU_MULTI_THREADING = True

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(self: SklearnRunnable, *args: t.Any, **kwargs: t.Any) -> t.Any:
            # TODO: set inner_max_num_threads and n_jobs param here base on strategy env vars
            with parallel_backend(backend="loky"):
                return getattr(self.model, method_name)(*args, **kwargs)

        SklearnRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return SklearnRunnable



from __future__ import annotations

import math
import os
import tarfile
import typing as t
from concurrent.futures import ThreadPoolExecutor
from tempfile import NamedTemporaryFile
from tempfile import mkstemp

import attrs
import fs
import httpx
from simple_di import Provide
from simple_di import inject

from ...exceptions import BentoMLException
from ...exceptions import NotFound
from ..bento import Bento
from ..bento import BentoStore
from ..configuration.containers import BentoMLContainer
from ..tag import Tag
from ..utils.filesystem import safe_extract_tarfile
from .base import FILE_CHUNK_SIZE
from .base import UPLOAD_RETRY_COUNT
from .base import CallbackIOWrapper
from .base import Spinner
from .model import ModelAPI
from .schemas.modelschemas import UploadStatus
from .schemas.schemasv1 import BentoSchema
from .schemas.schemasv1 import CompleteMultipartUploadSchema
from .schemas.schemasv1 import CompletePartSchema
from .schemas.schemasv1 import CreateBentoRepositorySchema
from .schemas.schemasv1 import CreateBentoSchema
from .schemas.schemasv1 import FinishUploadSchema
from .schemas.schemasv1 import LabelItemSchema
from .schemas.schemasv1 import PreSignMultipartUploadUrlSchema
from .schemas.schemasv1 import TransmissionStrategy
from .schemas.schemasv1 import UpdateBentoSchema

if t.TYPE_CHECKING:
    from concurrent.futures import Future

    from rich.progress import TaskID

    from _bentoml_sdk.models import Model

    from .client import RestApiClient
    from .schemas.schemasv1 import BentoWithRepositoryListSchema


@attrs.frozen
class BentoAPI:
    _client: RestApiClient = attrs.field(repr=False)
    spinner: Spinner = attrs.field(repr=False, factory=Spinner)

    def push(
        self,
        bento: Bento,
        *,
        force: bool = False,
        bare: bool = False,
        threads: int = 10,
    ) -> None:
        """Push a Bento to the remote Bento store

        Args:
            bento: The Bento to push
            force: Whether to force push the Bento
            bare: If true, only push the Bento manifest
            threads: The number of threads to use for the push
        """
        with self.spinner:
            upload_task_id = self.spinner.transmission_progress.add_task(
                f'Pushing Bento "{bento.tag}"', start=False, visible=False
            )
            self._do_push_bento(
                bento, upload_task_id, force=force, threads=threads, bare=bare
            )

    @inject
    def _do_push_bento(
        self,
        bento: Bento,
        upload_task_id: TaskID,
        *,
        force: bool = False,
        threads: int = 10,
        bare: bool = False,
        bentoml_tmp_dir: str = Provide[BentoMLContainer.tmp_bento_store_dir],
    ):
        rest_client = self._client
        from _bentoml_sdk.models import BentoModel
        from _bentoml_sdk.models import HuggingFaceModel

        name = bento.tag.name
        version = bento.tag.version
        if version is None:
            raise BentoMLException(f"Bento {bento.tag} version cannot be None")
        info = bento.info
        models_to_push: list[Model[t.Any]] = []
        for model in info.all_models:
            if model.registry == "huggingface":
                models_to_push.append(HuggingFaceModel.from_info(model))
            else:
                model = BentoModel.from_info(model)
                if model.stored is not None:
                    models_to_push.append(model)
        model_api = ModelAPI(self._client, self.spinner)

        def push_model(model: Model[t.Any]) -> None:
            model_upload_task_id = self.spinner.transmission_progress.add_task(
                f'Pushing model "{model}"', start=False, visible=False
            )
            model_api._do_push_model(
                model,
                model_upload_task_id,
                force=force,
                threads=threads,
            )

        with ThreadPoolExecutor(max_workers=max(len(models_to_push), 1)) as executor:
            list(executor.map(push_model, models_to_push))

        with self.spinner.spin(text=f'Fetching Bento repository "{name}"'):
            bento_repository = rest_client.v1.get_bento_repository(
                bento_repository_name=name
            )
        if not bento_repository:
            with self.spinner.spin(
                text=f'Bento repository "{name}" not found, creating now..'
            ):
                bento_repository = rest_client.v1.create_bento_repository(
                    req=CreateBentoRepositorySchema(name=name, description="")
                )
        with self.spinner.spin(
            text=f'Try fetching Bento "{bento.tag}" from remote Bento store..'
        ):
            remote_bento = rest_client.v1.get_bento(
                bento_repository_name=name, version=version
            )
        if (
            not force
            and remote_bento
            and remote_bento.upload_status == UploadStatus.SUCCESS.value
        ):
            self.spinner.log(
                f'[bold blue]Push skipped: Bento "{bento.tag}" already exists in remote Bento store'
            )
            return
        labels: list[LabelItemSchema] = [
            LabelItemSchema(key=key, value=value) for key, value in info.labels.items()
        ]
        manifest = bento.get_manifest(dev=bare)
        if not remote_bento:
            with self.spinner.spin(
                text=f'Registering Bento "{bento.tag}" with remote Bento store..'
            ):
                remote_bento = rest_client.v1.create_bento(
                    bento_repository_name=bento_repository.name,
                    req=CreateBentoSchema(
                        description=bento.doc,
                        version=version,
                        build_at=info.creation_time,
                        manifest=manifest,
                        labels=labels,
                    ),
                )
        else:
            with self.spinner.spin(text=f'Updating Bento "{bento.tag}"..'):
                remote_bento = rest_client.v1.update_bento(
                    bento_repository_name=bento_repository.name,
                    version=version,
                    req=UpdateBentoSchema(
                        description=bento.doc,
                        manifest=manifest,
                        labels=labels,
                    ),
                )
        if bare:
            with self.spinner.spin(
                text="Submitting upload status to remote Bento store"
            ):
                rest_client.v1.finish_upload_bento(
                    bento_repository_name=bento_repository.name,
                    version=version,
                    req=FinishUploadSchema(
                        status=UploadStatus.SUCCESS.value,
                        reason="bento for development",
                    ),
                )
            return
        transmission_strategy: TransmissionStrategy = "proxy"
        presigned_upload_url: str | None = None

        if remote_bento.transmission_strategy is not None:
            transmission_strategy = remote_bento.transmission_strategy
        else:
            with self.spinner.spin(
                text=f'Getting a presigned upload url for bento "{bento.tag}" ..'
            ):
                remote_bento = rest_client.v1.presign_bento_upload_url(
                    bento_repository_name=bento_repository.name, version=version
                )
                if remote_bento.presigned_upload_url:
                    transmission_strategy = "presigned_url"
                    presigned_upload_url = remote_bento.presigned_upload_url

        def io_cb(x: int):
            self.spinner.transmission_progress.update(upload_task_id, advance=x)

        fd, tar_name = mkstemp(
            prefix="bentoml-bento-", suffix=".tar", dir=bentoml_tmp_dir
        )
        tar_io = os.fdopen(fd, "wb+")
        try:
            with self.spinner.spin(
                text=f'Creating tar archive for bento "{bento.tag}"..'
            ):
                with tarfile.open(fileobj=tar_io, mode="w:") as tar:

                    def filter_(
                        tar_info: tarfile.TarInfo,
                    ) -> t.Optional[tarfile.TarInfo]:
                        if tar_info.path == "./models" or tar_info.path.startswith(
                            "./models/"
                        ):
                            return None
                        return tar_info

                    tar.add(bento.path, arcname="./", filter=filter_)

            with self.spinner.spin(text=f'Start uploading bento "{bento.tag}"..'):
                rest_client.v1.start_upload_bento(
                    bento_repository_name=bento_repository.name, version=version
                )
            file_size = tar_io.tell()
            io_with_cb = CallbackIOWrapper(tar_io, read_cb=io_cb)

            self.spinner.transmission_progress.update(
                upload_task_id, completed=0, total=file_size, visible=True
            )
            self.spinner.transmission_progress.start_task(upload_task_id)

            if transmission_strategy == "proxy":
                try:
                    rest_client.v1.upload_bento(
                        bento_repository_name=bento_repository.name,
                        version=version,
                        data=io_with_cb,
                    )
                except Exception as e:  # pylint: disable=broad-except
                    self.spinner.log(
                        f'[bold red]:police_car_light: Failed to upload bento "{bento.tag}"'
                    )
                    raise e
                self.spinner.log(
                    f'[bold green]:white_check_mark: Successfully pushed bento "{bento.tag}"'
                )
                return
            finish_req = FinishUploadSchema(
                status=UploadStatus.SUCCESS.value, reason=""
            )
            try:
                if presigned_upload_url is not None:
                    resp = httpx.put(
                        presigned_upload_url, content=io_with_cb, timeout=36000
                    )
                    if resp.status_code != 200:
                        finish_req = FinishUploadSchema(
                            status=UploadStatus.FAILED.value,
                            reason=resp.text,
                        )
                else:
                    with self.spinner.spin(
                        text=f'Start multipart uploading Bento "{bento.tag}"...'
                    ):
                        remote_bento = rest_client.v1.start_bento_multipart_upload(
                            bento_repository_name=bento_repository.name,
                            version=version,
                        )
                        if not remote_bento.upload_id:
                            raise BentoMLException(
                                f'Failed to start multipart upload for Bento "{bento.tag}", upload_id is empty'
                            )

                        upload_id: str = remote_bento.upload_id

                    chunks_count = math.ceil(file_size / FILE_CHUNK_SIZE)
                    tar_io.close()

                    def chunk_upload(
                        upload_id: str, chunk_number: int
                    ) -> FinishUploadSchema | tuple[str, int]:
                        with self.spinner.spin(
                            text=f'({chunk_number}/{chunks_count}) Presign multipart upload url of Bento "{bento.tag}"...'
                        ):
                            remote_bento = (
                                rest_client.v1.presign_bento_multipart_upload_url(
                                    bento_repository_name=bento_repository.name,
                                    version=version,
                                    req=PreSignMultipartUploadUrlSchema(
                                        upload_id=upload_id,
                                        part_number=chunk_number,
                                    ),
                                )
                            )
                        with self.spinner.spin(
                            text=f'({chunk_number}/{chunks_count}) Uploading chunk of Bento "{bento.tag}"...'
                        ):
                            with open(tar_name, "rb") as f:
                                chunk_io = CallbackIOWrapper(
                                    f,
                                    read_cb=io_cb,
                                    start=(chunk_number - 1) * FILE_CHUNK_SIZE,
                                    end=chunk_number * FILE_CHUNK_SIZE
                                    if chunk_number < chunks_count
                                    else None,
                                )

                                for i in range(UPLOAD_RETRY_COUNT):
                                    resp = httpx.put(
                                        remote_bento.presigned_upload_url,
                                        content=chunk_io,
                                        timeout=36000,
                                    )
                                    if resp.status_code == 200:
                                        break
                                    if i == UPLOAD_RETRY_COUNT - 1:
                                        return FinishUploadSchema(
                                            status=UploadStatus.FAILED.value,
                                            reason=resp.text,
                                        )
                                    else:  # retry and reset and update progress
                                        read = chunk_io.reset()
                                        self.spinner.transmission_progress.update(
                                            upload_task_id, advance=-read
                                        )
                                return resp.headers["ETag"], chunk_number

                    futures_: list[Future[FinishUploadSchema | tuple[str, int]]] = []

                    with ThreadPoolExecutor(
                        max_workers=min(max(chunks_count, 1), threads)
                    ) as executor:
                        for i in range(1, chunks_count + 1):
                            future = executor.submit(
                                chunk_upload,
                                upload_id,
                                i,
                            )
                            futures_.append(future)

                    parts: list[CompletePartSchema] = []

                    for future in futures_:
                        result = future.result()
                        if isinstance(result, FinishUploadSchema):
                            finish_req = result
                            break
                        else:
                            etag, chunk_number = result
                            parts.append(
                                CompletePartSchema(
                                    part_number=chunk_number,
                                    etag=etag,
                                )
                            )

                    with self.spinner.spin(
                        text=f'Completing multipart upload of Bento "{bento.tag}"...'
                    ):
                        remote_bento = rest_client.v1.complete_bento_multipart_upload(
                            bento_repository_name=bento_repository.name,
                            version=version,
                            req=CompleteMultipartUploadSchema(
                                upload_id=upload_id,
                                parts=parts,
                            ),
                        )

            except Exception as e:  # pylint: disable=broad-except
                finish_req = FinishUploadSchema(
                    status=UploadStatus.FAILED.value,
                    reason=str(e),
                )
            if finish_req.status == UploadStatus.FAILED.value:
                self.spinner.log(f'[bold red]Failed to upload Bento "{bento.tag}"')
            with self.spinner.spin(
                text="Submitting upload status to remote Bento store"
            ):
                rest_client.v1.finish_upload_bento(
                    bento_repository_name=bento_repository.name,
                    version=version,
                    req=finish_req,
                )
            if finish_req.status != UploadStatus.SUCCESS.value:
                self.spinner.log(
                    f'[bold red]Failed pushing Bento "{bento.tag}": {finish_req.reason}'
                )
                raise BentoMLException(f'Failed to upload Bento "{bento.tag}"')
            else:
                self.spinner.log(f'[bold green]Successfully pushed Bento "{bento.tag}"')
        finally:
            try:
                tar_io.close()
            except OSError:
                pass
            os.unlink(tar_name)

    @inject
    def pull(
        self,
        tag: str | Tag,
        *,
        force: bool = False,
        with_models: bool = False,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> Bento:
        """Pull a bento from remote bento store

        Args:
            tag: The tag of the bento to pull
            force: Whether to force pull the bento
            bento_store: The bento store to pull the bento to

        Returns:
            The pulled bento
        """
        with self.spinner:
            download_task_id = self.spinner.transmission_progress.add_task(
                f'Pulling bento "{tag}"', start=False, visible=False
            )
            return self._do_pull_bento(
                tag,
                download_task_id,
                force=force,
                with_models=with_models,
                bento_store=bento_store,
            )

    def _pull_bento_models(self, bento: Bento) -> None:
        for model in bento.info.all_models:
            model.to_model().resolve()

    def _do_pull_bento(
        self,
        tag: str | Tag,
        download_task_id: TaskID,
        *,
        force: bool = False,
        with_models: bool = False,
        bento_store: BentoStore = Provide[BentoMLContainer.bento_store],
    ) -> Bento:
        rest_client = self._client
        try:
            bento = bento_store.get(tag)
            if not force:
                self.spinner.log(
                    f'[bold blue]Bento "{tag}" exists in local bento store'
                )
                if with_models:
                    self._pull_bento_models(bento)
                return bento
            bento_store.delete(tag)
        except NotFound:
            pass
        _tag = Tag.from_taglike(tag)
        name = _tag.name
        version = _tag.version
        if version is None:
            raise BentoMLException(f'Bento "{_tag}" version can not be None')

        with self.spinner.spin(text=f'Fetching bento "{_tag}"'):
            remote_bento = rest_client.v1.get_bento(
                bento_repository_name=name, version=version
            )
        if not remote_bento:
            raise BentoMLException(f'Bento "{_tag}" not found on remote Bento store')

        # Download bento files from remote Bento store
        transmission_strategy: TransmissionStrategy = "proxy"
        presigned_download_url: str | None = None

        if remote_bento.transmission_strategy is not None:
            transmission_strategy = remote_bento.transmission_strategy
        else:
            with self.spinner.spin(
                text=f'Getting a presigned download url for bento "{_tag}"'
            ):
                remote_bento = rest_client.v1.presign_bento_download_url(name, version)
                if remote_bento.presigned_download_url:
                    presigned_download_url = remote_bento.presigned_download_url
                    transmission_strategy = "presigned_url"

        if transmission_strategy == "proxy":
            response_ctx = rest_client.v1.download_bento(
                bento_repository_name=name,
                version=version,
            )
        else:
            if presigned_download_url is None:
                with self.spinner.spin(
                    text=f'Getting a presigned download url for bento "{_tag}"'
                ):
                    remote_bento = rest_client.v1.presign_bento_download_url(
                        name, version
                    )
                    presigned_download_url = remote_bento.presigned_download_url

            response_ctx = httpx.stream("GET", presigned_download_url)

        with NamedTemporaryFile() as tar_file:
            with response_ctx as response:
                if response.status_code != 200:
                    response.read()
                    raise BentoMLException(
                        f'Failed to download bento "{_tag}": {response.text}'
                    )
                total_size_in_bytes = int(response.headers.get("content-length", 0))
                block_size = 1024  # 1 Kibibyte
                self.spinner.transmission_progress.update(
                    download_task_id,
                    completed=0,
                    total=total_size_in_bytes,
                    visible=True,
                )
                self.spinner.transmission_progress.start_task(download_task_id)
                for data in response.iter_bytes(block_size):
                    self.spinner.transmission_progress.update(
                        download_task_id, advance=len(data)
                    )
                    tar_file.write(data)

            self.spinner.log(
                f'[bold green]Finished downloading all bento "{_tag}" files'
            )
            tar_file.seek(0, 0)
            tar = tarfile.open(fileobj=tar_file, mode="r")
            with self.spinner.spin(text=f'Extracting bento "{_tag}" tar file'):
                with fs.open_fs("temp://") as temp_fs:
                    safe_extract_tarfile(tar, temp_fs.getsyspath("/"))
                    bento = Bento.from_fs(temp_fs)
                    bento = bento.save(bento_store)
                    self.spinner.log(f'[bold green]Successfully pulled bento "{_tag}"')
                    if with_models:
                        self._pull_bento_models(bento)
                    return bento

    def list(self) -> BentoWithRepositoryListSchema:
        """List all bentos in the remote bento store

        Returns:
            The list of bentos
        """
        res = self._client.v1.get_bentos_list()
        if res is None:
            raise BentoMLException("List bentos request failed")

        res.items = [
            bento
            for bento in sorted(res.items, key=lambda x: x.created_at, reverse=True)
        ]
        return res

    def get(self, name: str, version: str | None) -> BentoSchema:
        """Get a bento by name and version

        Args:
            name: The name of the bento
            version: The version of the bento

        Returns:
            The bento
        """
        rest_client = self._client
        if version is None or version == "latest":
            res = rest_client.v1.list_bentos(bento_repository_name=name)
            if res is None:
                raise BentoMLException("List bento request failed")
            return res.items[0]
        res = rest_client.v1.get_bento(bento_repository_name=name, version=version)
        if res is None:
            raise NotFound(f'Bento "{name}:{version}" not found')
        return res



from __future__ import annotations

import itertools
import logging
import typing as t
from typing import TYPE_CHECKING

import anyio

from bentoml.exceptions import InvalidArgument

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from ..runner.container import Payload


T = t.TypeVar("T", covariant=True)
To = t.TypeVar("To")
Ti = t.TypeVar("Ti")


CUDA_SUCCESS = 0


def pass_through(i: T) -> T:
    return i


class Params(t.Generic[T]):
    """
    A container for */** parameters. It helps to perform an operation on all the params
    values at the same time.
    """

    args: tuple[T, ...]
    kwargs: dict[str, T]

    def __init__(self, *args: T, **kwargs: T):
        self.args = args
        self.kwargs = kwargs

    def items(self) -> t.Iterator[t.Tuple[t.Union[int, str], T]]:
        return itertools.chain(enumerate(self.args), self.kwargs.items())

    @classmethod
    def from_dict(cls, data: dict[str | int, T]) -> Params[T]:
        return cls(
            *(data[k] for k in sorted(k for k in data if isinstance(k, int))),
            **{k: v for k, v in data.items() if isinstance(k, str)},
        )

    def all_equal(self) -> bool:
        value_iter = iter(self.items())
        _, first = next(value_iter)
        return all(v == first for _, v in value_iter)

    def map(self, function: t.Callable[[T], To]) -> Params[To]:
        """
        Apply a function to all the values in the Params and return a Params of the
        return values.
        """
        args = tuple(function(a) for a in self.args)
        kwargs = {k: function(v) for k, v in self.kwargs.items()}
        return Params[To](*args, **kwargs)

    def map_enumerate(
        self, function: t.Callable[[T, Ti], To], iterable: t.Iterable[Ti]
    ) -> Params[To]:
        """
        Apply function that takes two arguments with given iterable as index to all none empty field in Params.
        """
        if self.args:
            return Params[To](
                *tuple(function(a, b) for a, b in zip(self.args, iterable))
            )
        return Params[To](
            **{k: function(self.kwargs[k], b) for k, b in zip(self.kwargs, iterable)}
        )

    def iter(self: Params[t.Iterable[Ti]]) -> t.Iterator[Params[Ti]]:
        """
        Iter over a Params of iterable values into a list of Params. All values should
        have the same length.
        """

        iter_params = self.map(iter)
        try:
            while True:
                args = tuple(next(a) for a in iter_params.args)
                kwargs = {k: next(v) for k, v in iter_params.kwargs.items()}
                yield Params(*args, **kwargs)
        except StopIteration:
            pass

    @classmethod
    def agg(
        cls,
        params_list: t.Sequence[Params[Ti]],
        agg_func: t.Callable[[t.Sequence[Ti]], To] = pass_through,
    ) -> Params[To]:
        """
        Aggregate a list of Params into a single Params by performing the aggregate
        function on the list of values at the same position.
        """
        if not params_list:
            return Params()

        args = tuple(
            agg_func(tuple(params.args[i] for params in params_list))
            for i, _ in enumerate(params_list[0].args)
        )
        kwargs = {
            k: agg_func(tuple(params.kwargs[k] for params in params_list))
            for k in params_list[0].kwargs
        }
        return Params(*args, **kwargs)

    @property
    def sample(self) -> T:
        """
        Return a sample value (the first value of args or kwargs if args is empty)
        of the Params.
        """
        if self.args:
            return self.args[0]
        return next(iter(self.kwargs.values()))


PAYLOAD_META_HEADER = "Bento-Payload-Meta"


def payload_paramss_to_batch_params(
    paramss: t.Sequence[Params[Payload]],
    batch_dim: int,
    # TODO: support mapping from arg to batch dimension
) -> tuple[Params[t.Any], list[int]]:
    from ..runner.container import AutoContainer

    _converted_params = Params.agg(
        paramss,
        agg_func=lambda i: AutoContainer.from_batch_payloads(i, batch_dim),
    ).iter()
    batched_params = next(_converted_params)
    indice_params: Params[list[int]] = next(_converted_params)

    # considering skip this check if the CPU overhead of each inference is too high
    if not indice_params.all_equal():
        raise InvalidArgument(
            f"argument lengths for parameters do not match: {tuple(indice_params.items())}"
        )
    return batched_params, indice_params.sample


# This is a reference to starlette.concurrency
class _StopIteration(Exception):
    pass


def _next(iterator: t.Iterator[T]) -> T:
    # We can't raise `StopIteration` from within the threadpool iterator
    # and catch it outside that context, so we coerce them into a different
    # exception type.
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(
    iterator: t.Iterator[T],
) -> t.AsyncIterator[T]:
    while True:
        try:
            yield await anyio.to_thread.run_sync(_next, iterator)
        except _StopIteration:
            break



from __future__ import annotations

import functools
import inspect
import sys
import typing as t

import attrs
import pydantic

from bentoml._internal.service.openapi import SUCCESS_DESCRIPTION
from bentoml._internal.service.openapi.specification import Encoding
from bentoml._internal.service.openapi.specification import MediaType
from bentoml._internal.service.openapi.specification import Schema
from bentoml._internal.utils import dict_filter_none

from .io_models import IODescriptor
from .io_models import IOMixin
from .io_models import ensure_io_descriptor

R = t.TypeVar("R")
T = t.TypeVar("T", bound="APIMethod[..., t.Any]")
if t.TYPE_CHECKING:
    P = t.ParamSpec("P")
else:
    P = t.TypeVar("P")


DEFAULT_STREAM_MEDIA_TYPE = "text/event-stream"


def _only_include(data: dict[str, t.Any], fields: t.Container[str]) -> dict[str, t.Any]:
    return {k: v for k, v in data.items() if k in fields}


def _io_descriptor_converter(it: t.Any) -> type[IODescriptor]:
    if not inspect.isclass(it):
        raise ValueError(f"{it} must be a class type")
    if not issubclass(it, (pydantic.BaseModel, dict)):
        raise ValueError(f"{it} is not a valid IODescriptor accepted type.")
    if issubclass(it, IOMixin):
        return it
    return ensure_io_descriptor(it)


@attrs.define
class APIMethod(t.Generic[P, R]):
    func: t.Callable[t.Concatenate[t.Any, P], R]
    route: str = attrs.field()
    name: str = attrs.field(init=False)
    input_spec: type[IODescriptor] = attrs.field(converter=_io_descriptor_converter)
    output_spec: type[IODescriptor] = attrs.field(converter=_io_descriptor_converter)
    batchable: bool = False
    batch_dim: tuple[int, int] = attrs.field(
        default=(0, 0), converter=lambda x: (x, x) if not isinstance(x, tuple) else x
    )
    max_batch_size: int = attrs.field(default=100, validator=attrs.validators.gt(1))
    max_latency_ms: int = 60000
    is_stream: bool = attrs.field(init=False)
    doc: str | None = attrs.field(init=False)
    ctx_param: str | None = attrs.field(init=False)
    is_task: bool = False

    @doc.default
    def default_doc(self) -> str | None:
        return self.func.__doc__

    @ctx_param.default
    def default_ctx_param(self) -> str | None:
        from bentoml import Context

        from .io_models import eval_type_lenient

        try:
            module = sys.modules[self.func.__module__]
        except KeyError:
            ns = None
        else:
            ns = module.__dict__

        parameters = inspect.signature(self.func).parameters
        if "ctx" in parameters:
            annotation = parameters["ctx"].annotation
            if (
                annotation is inspect.Parameter.empty
                or eval_type_lenient(annotation, ns, None) is Context
            ):
                return "ctx"
        elif "context" in parameters:
            annotation = parameters["context"].annotation
            if (
                annotation is inspect.Parameter.empty
                or eval_type_lenient(annotation, ns, None) is Context
            ):
                return "context"
        return None

    @is_stream.default
    def default_is_stream(self) -> bool:
        return inspect.isasyncgenfunction(self.func) or inspect.isgeneratorfunction(
            self.func
        )

    @name.default
    def default_name(self) -> str:
        return self.func.__name__

    @route.default
    def default_route(self) -> str:
        if self.func.__name__ == "__call__":
            return "/"
        return f"/{self.func.__name__}"

    @input_spec.default
    def default_input_spec(self) -> type[IODescriptor]:
        ctx_param = self.default_ctx_param()
        return IODescriptor.from_input(
            self.func,
            skip_self=True,
            skip_names=(ctx_param,) if ctx_param else (),
        )

    @output_spec.default
    def default_output_spec(self) -> type[IODescriptor]:
        return IODescriptor.from_output(self.func)

    def __attrs_post_init__(self) -> None:
        if self.is_stream and not self.output_spec.media_type:
            self.output_spec.media_type = DEFAULT_STREAM_MEDIA_TYPE

    @t.overload
    def __get__(self: T, instance: None, owner: type) -> T: ...

    @t.overload
    def __get__(self, instance: object, owner: type) -> t.Callable[P, R]: ...

    def __get__(self: T, instance: t.Any, owner: type) -> t.Callable[P, R] | T:
        if instance is None:
            return self

        local_caller = self._local_call(instance)

        if proxy := getattr(instance, "__self_proxy__", None):
            func = getattr(proxy, self.name)
        else:
            func = local_caller
        func.local = local_caller  # type: ignore[attr-defined]
        return func

    def _local_call(self, instance: t.Any) -> t.Callable[P, R]:
        from pydantic.fields import FieldInfo
        from pydantic_core import PydanticUndefined

        func_sig = inspect.signature(self.func)
        # skip the `self` parameter
        params = list(func_sig.parameters.values())[1:]

        @functools.wraps(self.func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            # Extract the default values from `Field` if any
            for i, param in enumerate(params):
                if i < len(args) or param.name in kwargs:
                    # skip the arguments that are already provided
                    continue
                if isinstance(field := param.default, FieldInfo):
                    if field.default is not PydanticUndefined:
                        kwargs[param.name] = field.default
                    elif field.default_factory not in (None, PydanticUndefined):
                        kwargs[param.name] = field.default_factory()
            return self.func(instance, *args, **kwargs)

        wrapped.__signature__ = func_sig.replace(parameters=params)
        wrapped.__is_bentoml_api_func__ = True
        # same as functools.partial in order that inspect can recognize it
        wrapped.func = self.func
        return wrapped

    def schema(self) -> dict[str, t.Any]:
        output = _flatten_model_schema(self.output_spec)
        if self.is_stream:
            output["is_stream"] = True
        if self.output_spec.media_type:
            output["media_type"] = self.output_spec.media_type
        return dict_filter_none(
            {
                "name": self.name,
                "route": self.route,
                "description": self.__doc__,
                "batchable": self.batchable,
                "input": _flatten_model_schema(self.input_spec),
                "output": output,
                "is_task": self.is_task,
            }
        )

    def openapi_request(self) -> dict[str, t.Any]:
        from .service.openapi import REF_TEMPLATE

        input = _flatten_field(
            _only_include(
                self.input_spec.model_json_schema(
                    ref_template=REF_TEMPLATE, mode="serialization"
                ),
                [attr.name for attr in Schema.__attrs_attrs__],
            ),
            {},
            max_depth=1,
        )
        encoding: dict[str, Encoding] = {}
        if self.input_spec.mime_type() == "multipart/form-data":
            for field in self.input_spec.model_fields:
                if field not in self.input_spec.multipart_fields:
                    encoding[field] = Encoding(contentType="application/json")

        return {
            "content": {
                self.input_spec.mime_type(): MediaType(
                    schema=Schema(**input), encoding=encoding or None
                )
            },
        }

    def openapi_response(self) -> dict[str, t.Any]:
        from .service.openapi import REF_TEMPLATE

        output = _flatten_field(
            _only_include(
                self.output_spec.model_json_schema(
                    ref_template=REF_TEMPLATE, mode="serialization"
                ),
                [attr.name for attr in Schema.__attrs_attrs__],
            ),
            {},
            max_depth=1,
        )

        return {
            "description": SUCCESS_DESCRIPTION,
            "content": {
                self.output_spec.mime_type(): MediaType(schema=Schema(**output))
            },
        }


def _flatten_field(
    schema: dict[str, t.Any],
    defs: dict[str, t.Any],
    max_depth: int | None = None,
    _depth: int = 0,
) -> dict[str, t.Any]:
    if "allOf" in schema:
        schema.update(schema.pop("allOf")[0])
    if "anyOf" in schema:
        schema.update(schema.pop("anyOf")[0])
    if max_depth is not None and _depth >= max_depth:
        return schema
    if "$ref" in schema:
        ref = schema.pop("$ref")[len("#/$defs/") :]
        schema.update(_flatten_field(defs[ref], defs, max_depth, _depth=_depth + 1))
    elif schema.get("type") == "object" and "properties" in schema:
        for k, v in schema["properties"].items():
            schema["properties"][k] = _flatten_field(
                v, defs, max_depth, _depth=_depth + 1
            )
    elif schema.get("type") == "array" and "items" in schema:
        schema["items"] = _flatten_field(
            schema["items"], defs, max_depth, _depth=_depth + 1
        )
    return schema


def _flatten_model_schema(model: type[IODescriptor]) -> dict[str, t.Any]:
    schema = model.model_json_schema()
    if not schema.get("properties"):
        return schema
    defs = schema.pop("$defs", {})
    return _flatten_field(schema, defs)



from __future__ import annotations

import typing as t

from pydantic._internal import _known_annotated_metadata
from pydantic._internal._typing_extra import is_annotated

from .typing_utils import get_args
from .typing_utils import get_origin
from .validators import ContentType
from .validators import DataframeSchema
from .validators import DType
from .validators import FileSchema
from .validators import PILImageEncoder
from .validators import Shape
from .validators import TensorSchema

if t.TYPE_CHECKING:
    from pydantic import ConfigDict

    Preparer: t.TypeAlias = t.Callable[
        [t.Any, t.Iterable[t.Any], ConfigDict], tuple[t.Any, list[t.Any]] | None
    ]
    import numpy as np
    import pandas as pd
    import tensorflow as tf
    import torch
    from PIL import Image
else:
    from bentoml._internal.utils.lazy_loader import LazyLoader

    np = LazyLoader("np", globals(), "numpy")
    tf = LazyLoader("tf", globals(), "tensorflow")
    torch = LazyLoader("torch", globals(), "torch")
    pd = LazyLoader("pd", globals(), "pandas")
    Image = LazyLoader("Image", globals(), "PIL.Image")


def numpy_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    if not getattr(source, "__module__", "").startswith("numpy"):
        return None

    origin = get_origin(source)

    if origin is not np.ndarray:
        return None

    args = get_args(source)
    try:
        dtype = np.dtype(get_args(args[1])[0]).name if args else None
    except TypeError:
        return None
    shape: tuple[int, ...] | None = None

    _, other_annotations = _known_annotated_metadata.collect_known_metadata(annotations)
    remaining_annotations = []
    for annotation in other_annotations:
        if isinstance(annotation, Shape):
            shape = annotation.dimensions
        elif isinstance(annotation, DType):
            if dtype is not None and dtype != annotation.dtype:
                raise ValueError(
                    f"Conflicting dtype annotations: {dtype} and {annotation.dtype}"
                )
            dtype = annotation.dtype
        else:
            remaining_annotations.append(annotation)
    return source, [TensorSchema("numpy-array", dtype, shape), *remaining_annotations]


def torch_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    if not getattr(source, "__module__", "").startswith("torch"):
        return None

    origin = get_origin(source)

    if not issubclass(origin, torch.Tensor):
        return None

    dtype: str | None = None
    shape: tuple[int, ...] | None = None

    _, remaining_annotations = _known_annotated_metadata.collect_known_metadata(
        annotations
    )
    for i, annotation in reversed(list(enumerate(remaining_annotations))):
        if isinstance(annotation, Shape):
            shape = annotation.dimensions
            del remaining_annotations[i]
        elif isinstance(annotation, DType):
            dtype = annotation.dtype
            del remaining_annotations[i]
    return source, [TensorSchema("torch-tensor", dtype, shape), *remaining_annotations]


def tf_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    if not getattr(source, "__module__", "").startswith("tensorflow"):
        return None

    origin = get_origin(source)
    if not issubclass(origin, tf.Tensor):
        return None

    dtype: str | None = None
    shape: tuple[int, ...] | None = None

    _, remaining_annotations = _known_annotated_metadata.collect_known_metadata(
        annotations
    )
    for i, annotation in reversed(list(enumerate(remaining_annotations))):
        if isinstance(annotation, Shape):
            shape = annotation.dimensions
            del remaining_annotations[i]
        elif isinstance(annotation, DType):
            dtype = annotation.dtype
            del remaining_annotations[i]
    return source, [TensorSchema("tf-tensor", dtype, shape), *remaining_annotations]


def pandas_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    if not getattr(source, "__module__", "").startswith("pandas"):
        return None

    origin = get_origin(source)
    if not issubclass(origin, pd.DataFrame):
        return None

    _, remaining_annotations = _known_annotated_metadata.collect_known_metadata(
        annotations
    )
    if not any(isinstance(a, DataframeSchema) for a in remaining_annotations):
        remaining_annotations.insert(0, DataframeSchema())
    return origin, remaining_annotations


def pil_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], _config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    if not getattr(source, "__module__", "").startswith("PIL."):
        return None

    origin = get_origin(source) or source
    if not issubclass(origin, Image.Image):
        return None

    _, remaining_annotations = _known_annotated_metadata.collect_known_metadata(
        annotations
    )
    return origin, [PILImageEncoder(), *remaining_annotations]


def pathlib_prepare_pydantic_annotations(
    source: t.Any, annotations: t.Iterable[t.Any], _config: ConfigDict
) -> tuple[t.Any, list[t.Any]] | None:
    import pathlib

    if source not in {
        pathlib.PurePath,
        pathlib.PurePosixPath,
        pathlib.PureWindowsPath,
        pathlib.Path,
        pathlib.PosixPath,
        pathlib.WindowsPath,
    }:
        return None

    _, remaining_annotations = _known_annotated_metadata.collect_known_metadata(
        annotations
    )
    content_type: str | None = None
    for i, annotation in reversed(list(enumerate(remaining_annotations))):
        if isinstance(annotation, ContentType):
            content_type = annotation.content_type
            del remaining_annotations[i]
    return source, [FileSchema(content_type=content_type), *remaining_annotations]


CUSTOM_PREPARE_METHODS = [
    # path
    pathlib_prepare_pydantic_annotations,
    # tensors
    numpy_prepare_pydantic_annotations,
    torch_prepare_pydantic_annotations,
    tf_prepare_pydantic_annotations,
    # dataframe
    pandas_prepare_pydantic_annotations,
    # PIL image
    pil_prepare_pydantic_annotations,
]

SUPPORTED_CONTAINER_TYPES = [
    t.Union,
    list,
    t.List,
    dict,
    t.Dict,
    t.AsyncGenerator,
    t.AsyncIterable,
    t.AsyncIterator,
    t.Generator,
    t.Iterable,
    t.Iterator,
]


def patch_annotation(
    annotation: t.Any, model_config: ConfigDict, metadata: t.Iterable[t.Any] = ()
) -> tuple[t.Any, list[t.Any]]:
    import typing_extensions as te

    origin, args = te.get_origin(annotation), te.get_args(annotation)
    if origin in SUPPORTED_CONTAINER_TYPES:
        patched_args: list[t.Any] = []
        for arg in args:
            ann, extras = patch_annotation(arg, model_config)
            if extras:
                ann = te.Annotated[(ann, *extras)]
            patched_args.append(ann)
        return origin[tuple(patched_args)], list(metadata)

    if is_annotated(annotation):
        source, *annotations = args
        metadata = [*annotations, *metadata]
    else:
        source = annotation
        metadata = list(metadata)
    for method in CUSTOM_PREPARE_METHODS:
        result = method(source, metadata, model_config)
        if result is None:
            continue
        return result
    return annotation, list(metadata)



from __future__ import annotations

import math
import os
import tarfile
import typing as t
import warnings
from concurrent.futures import ThreadPoolExecutor
from tempfile import NamedTemporaryFile
from tempfile import mkstemp
from threading import Lock

import attrs
import fs
import httpx
from simple_di import Provide
from simple_di import inject

from ...exceptions import BentoMLException
from ...exceptions import NotFound
from ..configuration.containers import BentoMLContainer
from ..models import Model as StoredModel
from ..models import ModelStore
from ..tag import Tag
from ..utils.filesystem import safe_extract_tarfile
from .base import FILE_CHUNK_SIZE
from .base import UPLOAD_RETRY_COUNT
from .base import CallbackIOWrapper
from .base import Spinner
from .schemas.modelschemas import UploadStatus
from .schemas.schemasv1 import CompleteMultipartUploadSchema
from .schemas.schemasv1 import CompletePartSchema
from .schemas.schemasv1 import CreateModelRepositorySchema
from .schemas.schemasv1 import FinishUploadSchema
from .schemas.schemasv1 import ModelSchema
from .schemas.schemasv1 import PreSignMultipartUploadUrlSchema
from .schemas.schemasv1 import TransmissionStrategy

if t.TYPE_CHECKING:
    from concurrent.futures import Future

    from rich.progress import TaskID

    from _bentoml_sdk.models import Model

    from .client import RestApiClient
    from .schemas.schemasv1 import ModelWithRepositoryListSchema


@attrs.frozen
class ModelAPI:
    _client: RestApiClient = attrs.field(repr=False)
    spinner: Spinner = attrs.field(repr=False, factory=Spinner)
    _lock: Lock = attrs.field(repr=False, init=False, factory=Lock)

    def push(
        self,
        model: Model[t.Any],
        *,
        force: bool = False,
        threads: int = 10,
    ) -> None:
        """Push a model to remote model store

        Args:
            model: The model to push
            force: Whether to force push the model
            threads: The number of threads to use for the push
        """
        with self.spinner:
            upload_task_id = self.spinner.transmission_progress.add_task(
                f'Pushing model "{model}"', start=False, visible=False
            )
            self._do_push_model(model, upload_task_id, force=force, threads=threads)

    @inject
    def _do_push_model(
        self,
        model: Model[t.Any],
        upload_task_id: TaskID,
        *,
        force: bool = False,
        threads: int = 10,
        bentoml_tmp_dir: str = Provide[BentoMLContainer.tmp_bento_store_dir],
    ):
        from _bentoml_sdk.models import BentoModel

        rest_client = self._client

        model_info = model.to_info()
        name = model_info.tag.name
        version = model_info.tag.version
        if version is None:
            raise BentoMLException(f'Model "{model}" version cannot be None')

        with self._lock:
            # Models might be pushed by multiple threads at the same time
            # when they are under the same model repository, race condition
            # might happen when creating the model repository. So we need to
            # protect it with a lock.
            with self.spinner.spin(text=f'Fetching model repository "{name}"'):
                model_repository = rest_client.v1.get_model_repository(
                    model_repository_name=name
                )
            if not model_repository:
                with self.spinner.spin(
                    text=f'Model repository "{name}" not found, creating now..'
                ):
                    model_repository = rest_client.v1.create_model_repository(
                        req=CreateModelRepositorySchema(name=name, description="")
                    )
        with self.spinner.spin(
            text=f'Try fetching model "{model}" from remote model store..'
        ):
            remote_model = rest_client.v1.get_model(
                model_repository_name=name, version=version
            )
        if (
            not force
            and remote_model
            and remote_model.upload_status == UploadStatus.SUCCESS.value
        ):
            self.spinner.log(
                f'[bold blue]Model "{model}" already exists in remote model store, skipping'
            )
            return
        if not remote_model:
            with self.spinner.spin(
                text=f'Registering model "{model}" with remote model store..'
            ):
                remote_model = rest_client.v1.create_model(
                    model_repository_name=model_repository.name,
                    req=model.to_create_schema(),
                )
        if not isinstance(model, BentoModel):
            self.spinner.log(f"[bold blue]Skip uploading non-bentoml model {model}")
            return
        assert model.stored is not None
        transmission_strategy: TransmissionStrategy = "proxy"
        presigned_upload_url: str | None = None

        if remote_model.transmission_strategy is not None:
            transmission_strategy = remote_model.transmission_strategy
        else:
            with self.spinner.spin(
                text=f'Getting a presigned upload url for Model "{model.tag}" ..'
            ):
                remote_model = rest_client.v1.presign_model_upload_url(
                    model_repository_name=model_repository.name, version=version
                )
                if remote_model.presigned_upload_url:
                    transmission_strategy = "presigned_url"
                    presigned_upload_url = remote_model.presigned_upload_url

        def io_cb(x: int):
            self.spinner.transmission_progress.update(upload_task_id, advance=x)

        fd, tar_name = mkstemp(
            prefix="bentoml-model-", suffix=".tar", dir=bentoml_tmp_dir
        )
        tar_io = os.fdopen(fd, "wb+")
        try:
            with self.spinner.spin(
                text=f'Creating tar archive for model "{model.tag}"..'
            ):
                with tarfile.open(fileobj=tar_io, mode="w:") as tar:
                    tar.add(model.stored.path, arcname="./")
            with self.spinner.spin(text=f'Start uploading model "{model.tag}"..'):
                rest_client.v1.start_upload_model(
                    model_repository_name=model_repository.name, version=version
                )
            file_size = tar_io.tell()
            self.spinner.transmission_progress.update(
                upload_task_id,
                description=f'Uploading model "{model.tag}"',
                total=file_size,
                visible=True,
            )
            self.spinner.transmission_progress.start_task(upload_task_id)
            io_with_cb = CallbackIOWrapper(tar_io, read_cb=io_cb)

            if transmission_strategy == "proxy":
                try:
                    rest_client.v1.upload_model(
                        model_repository_name=model_repository.name,
                        version=version,
                        data=io_with_cb,
                    )
                except Exception as e:  # pylint: disable=broad-except
                    self.spinner.log(
                        f'[bold red]:police_car_light: Failed to upload model "{model.tag}"'
                    )
                    raise e
                self.spinner.log(
                    f'[bold green]:white_check_mark: Successfully pushed model "{model.tag}"'
                )
                return
            finish_req = FinishUploadSchema(
                status=UploadStatus.SUCCESS.value, reason=""
            )
            try:
                if presigned_upload_url is not None:
                    resp = httpx.put(
                        presigned_upload_url, content=io_with_cb, timeout=36000
                    )
                    if resp.status_code != 200:
                        finish_req = FinishUploadSchema(
                            status=UploadStatus.FAILED.value,
                            reason=resp.text,
                        )
                else:
                    with self.spinner.spin(
                        text=f'Start multipart uploading Model "{model.tag}"...'
                    ):
                        remote_model = rest_client.v1.start_model_multipart_upload(
                            model_repository_name=model_repository.name,
                            version=version,
                        )
                        if not remote_model.upload_id:
                            raise BentoMLException(
                                f'Failed to start multipart upload for model "{model.tag}", upload_id is empty'
                            )

                        upload_id: str = remote_model.upload_id

                    chunks_count = math.ceil(file_size / FILE_CHUNK_SIZE)
                    tar_io.close()

                    def chunk_upload(
                        upload_id: str, chunk_number: int
                    ) -> FinishUploadSchema | tuple[str, int]:
                        with self.spinner.spin(
                            text=f'({chunk_number}/{chunks_count}) Presign multipart upload url of model "{model.tag}"...'
                        ):
                            remote_model = (
                                rest_client.v1.presign_model_multipart_upload_url(
                                    model_repository_name=model_repository.name,
                                    version=version,
                                    req=PreSignMultipartUploadUrlSchema(
                                        upload_id=upload_id,
                                        part_number=chunk_number,
                                    ),
                                )
                            )

                        with self.spinner.spin(
                            text=f'({chunk_number}/{chunks_count}) Uploading chunk of model "{model.tag}"...'
                        ):
                            with open(tar_name, "rb") as f:
                                chunk_io = CallbackIOWrapper(
                                    f,
                                    read_cb=io_cb,
                                    start=(chunk_number - 1) * FILE_CHUNK_SIZE,
                                    end=chunk_number * FILE_CHUNK_SIZE
                                    if chunk_number < chunks_count
                                    else None,
                                )

                                for i in range(UPLOAD_RETRY_COUNT):
                                    resp = httpx.put(
                                        remote_model.presigned_upload_url,
                                        content=chunk_io,
                                        timeout=36000,
                                    )
                                    if resp.status_code == 200:
                                        break
                                    if i == UPLOAD_RETRY_COUNT - 1:
                                        return FinishUploadSchema(
                                            status=UploadStatus.FAILED.value,
                                            reason=resp.text,
                                        )
                                    else:  # retry and reset and update progress
                                        read = chunk_io.reset()
                                        self.spinner.transmission_progress.update(
                                            upload_task_id, advance=-read
                                        )
                                return resp.headers["ETag"], chunk_number

                    futures_: list[Future[FinishUploadSchema | tuple[str, int]]] = []

                    with ThreadPoolExecutor(
                        max_workers=min(max(chunks_count, 1), threads)
                    ) as executor:
                        for i in range(1, chunks_count + 1):
                            future = executor.submit(
                                chunk_upload,
                                upload_id,
                                i,
                            )
                            futures_.append(future)

                    parts: list[CompletePartSchema] = []

                    for future in futures_:
                        result = future.result()
                        if isinstance(result, FinishUploadSchema):
                            finish_req = result
                            break
                        else:
                            etag, chunk_number = result
                            parts.append(
                                CompletePartSchema(
                                    part_number=chunk_number,
                                    etag=etag,
                                )
                            )

                    with self.spinner.spin(
                        text=f'Completing multipart upload of model "{model.tag}"...'
                    ):
                        remote_model = rest_client.v1.complete_model_multipart_upload(
                            model_repository_name=model_repository.name,
                            version=version,
                            req=CompleteMultipartUploadSchema(
                                upload_id=upload_id,
                                parts=parts,
                            ),
                        )

            except Exception as e:  # pylint: disable=broad-except
                finish_req = FinishUploadSchema(
                    status=UploadStatus.FAILED.value,
                    reason=str(e),
                )
            if finish_req.status == UploadStatus.FAILED.value:
                self.spinner.log(f'[bold red]Failed to upload model "{model.tag}"')
            with self.spinner.spin(
                text="Submitting upload status to remote model store"
            ):
                rest_client.v1.finish_upload_model(
                    model_repository_name=model_repository.name,
                    version=version,
                    req=finish_req,
                )

            if finish_req.status != UploadStatus.SUCCESS.value:
                self.spinner.log(
                    f'[bold red]Failed pushing model "{model.tag}" : {finish_req.reason}'
                )
                raise BentoMLException(f'Failed to upload model "{model.tag}"')
            else:
                self.spinner.log(f'[bold green]Successfully pushed model "{model.tag}"')
        finally:
            try:
                tar_io.close()
            except OSError:
                pass
            os.unlink(tar_name)

    @inject
    def pull(
        self,
        tag: str | Tag,
        *,
        force: bool = False,
        model_store: ModelStore = Provide[BentoMLContainer.model_store],
        query: str | None = None,
    ) -> StoredModel | None:
        """Pull a model from remote model store

        Args:
            tag: The tag of the model to pull
            force: Whether to force pull the model
            model_store: The model store to pull the model to
            query: The query to use for the pull

        Returns:
            The pulled model
        """
        with self.spinner:
            download_task_id = self.spinner.transmission_progress.add_task(
                f'Pulling model "{tag}"', start=False, visible=False
            )
            return self._do_pull_model(
                tag,
                download_task_id,
                force=force,
                model_store=model_store,
                query=query,
            )

    @inject
    def _do_pull_model(
        self,
        tag: str | Tag,
        download_task_id: TaskID,
        *,
        force: bool = False,
        model_store: ModelStore = Provide[BentoMLContainer.model_store],
        query: str | None = None,
    ) -> StoredModel | None:
        rest_client = self._client
        _tag = Tag.from_taglike(tag)
        try:
            model = model_store.get(_tag)
        except NotFound:
            model = None
        else:
            if _tag.version not in (None, "latest"):
                if not force:
                    self.spinner.log(
                        f'[bold blue]Model "{tag}" already exists locally, skipping'
                    )
                    return model
                else:
                    model_store.delete(tag)
        name = _tag.name
        version = _tag.version
        if version in (None, "latest"):
            latest_model = rest_client.v1.get_latest_model(name, query=query)
            if latest_model is None:
                raise BentoMLException(
                    f'Model "{_tag}" not found on remote model store, you may need to specify a version'
                )
            if model is not None:
                if not force and latest_model.build_at < model.creation_time:
                    self.spinner.log(
                        f'[bold blue]Newer version of model "{name}" exists locally, skipping'
                    )
                    return model
                if model.tag.version == latest_model.version:
                    if not force:
                        self.spinner.log(
                            f'[bold blue]Model "{model.tag}" already exists locally, skipping'
                        )
                        return model
                    else:
                        model_store.delete(model.tag)
            version = latest_model.version
        elif query:
            warnings.warn(
                "`query` is ignored when model version is specified", UserWarning
            )

        with self.spinner.spin(
            text=f'Getting a presigned download url for model "{_tag}"..'
        ):
            remote_model = rest_client.v1.presign_model_download_url(name, version)

        if not remote_model:
            raise BentoMLException(f'Model "{_tag}" not found on remote model store')
        if remote_model.manifest.metadata.get("registry") == "huggingface":
            self.spinner.log(f"[bold blue]No content to download for model {_tag}")
            return
        # Download model files from remote model store
        transmission_strategy: TransmissionStrategy = "proxy"
        presigned_download_url: str | None = None

        if remote_model.transmission_strategy is not None:
            transmission_strategy = remote_model.transmission_strategy
        else:
            with self.spinner.spin(
                text=f'Getting a presigned download url for model "{_tag}"'
            ):
                remote_model = rest_client.v1.presign_model_download_url(name, version)
                if remote_model.presigned_download_url:
                    presigned_download_url = remote_model.presigned_download_url
                    transmission_strategy = "presigned_url"

        if transmission_strategy == "proxy":
            response_ctx = rest_client.v1.download_model(
                model_repository_name=name, version=version
            )
        else:
            if presigned_download_url is None:
                with self.spinner.spin(
                    text=f'Getting a presigned download url for model "{_tag}"'
                ):
                    remote_model = rest_client.v1.presign_model_download_url(
                        name, version
                    )
                    presigned_download_url = remote_model.presigned_download_url

            response_ctx = httpx.stream("GET", presigned_download_url)

        with NamedTemporaryFile() as tar_file:
            with response_ctx as response:
                if response.status_code != 200:
                    response.read()
                    raise BentoMLException(
                        f'Failed to download model "{_tag}": {response.text}'
                    )

                total_size_in_bytes = int(response.headers.get("content-length", 0))
                block_size = 1024  # 1 Kibibyte
                self.spinner.transmission_progress.update(
                    download_task_id,
                    description=f'Downloading model "{_tag}"',
                    total=total_size_in_bytes,
                    visible=True,
                )
                self.spinner.transmission_progress.start_task(download_task_id)
                for data in response.iter_bytes(block_size):
                    self.spinner.transmission_progress.update(
                        download_task_id, advance=len(data)
                    )
                    tar_file.write(data)

            self.spinner.log(f'[bold green]Finished downloading model "{_tag}" files')
            tar_file.seek(0, 0)
            tar = tarfile.open(fileobj=tar_file, mode="r")
            with self.spinner.spin(text=f'Extracting model "{_tag}" tar file'):
                with fs.open_fs("temp://") as temp_fs:
                    safe_extract_tarfile(tar, temp_fs.getsyspath("/"))
                    model = StoredModel.from_fs(temp_fs).save(model_store)
                    self.spinner.log(f'[bold green]Successfully pulled model "{_tag}"')
                    return model

    def list(self) -> ModelWithRepositoryListSchema:
        """List all models in the remote model store

        Returns:
            The list of models
        """
        res = self._client.v1.get_models_list()
        if res is None:
            raise BentoMLException("List models request failed")

        res.items = [
            model
            for model in sorted(res.items, key=lambda x: x.created_at, reverse=True)
        ]
        return res

    def get(self, name: str, version: str | None = None) -> ModelSchema:
        """Get a model from the remote model store

        Args:
            tag: The tag of the model to get

        Returns:
            The model
        """
        if version is None or version == "latest":
            res = self._client.v1.get_latest_model(name)
        else:
            res = self._client.v1.get_model(name, version)
        if res is None:
            raise NotFound(f'Model "{name}:{version}" not found')
        return res



from __future__ import annotations

import collections
import contextvars
import logging
import typing as t

MON_COLUMN_VAR: contextvars.ContextVar["dict[str, dict[str, str]] | None"] = (
    contextvars.ContextVar("MON_COLUMN_VAR", default=None)
)
MON_DATAS_VAR: contextvars.ContextVar["dict[str, collections.deque[t.Any]] | None"] = (
    contextvars.ContextVar("MON_DATAS_VAR", default=None)
)

DT = t.TypeVar("DT")
MT = t.TypeVar("MT", bound="MonitorBase[t.Any]")

BENTOML_MONITOR_ROLES = {"feature", "prediction", "target"}
BENTOML_MONITOR_TYPES = {"numerical", "categorical", "numerical_sequence"}


logger = logging.getLogger(__name__)


class MonitorBase(t.Generic[DT]):
    """
    The base monitor class. All monitors should inherit from this class.
    Subclasses should implement the following methods:
    - export_schema
    - export_data
    to export the schema and data to the desired format.
    """

    PRESERVED_COLUMNS: tuple[str, ...] = ()

    def __init__(
        self,
        name: str,
        **_: t.Any,
    ) -> None:
        self.name = name
        self.columns_schema: dict[str, dict[str, str]] | None = None

    def start_record(self):
        """
        Start recording data. This method should be called before logging any data.
        """
        if self.columns_schema is None:
            assert MON_COLUMN_VAR.get() is None
            MON_COLUMN_VAR.set({})

        assert MON_DATAS_VAR.get() is None
        MON_DATAS_VAR.set(collections.defaultdict(collections.deque))

    def stop_record(self) -> None:
        """
        Stop recording data. This method should be called after logging all data.
        """
        datas: dict[str, collections.deque[DT]] = MON_DATAS_VAR.get()  # type: ignore
        assert datas is not None

        if len(datas) == 0:
            logger.warning("No data logged in this record. Will skip output.")
            return

        if self.columns_schema is None:
            columns = MON_COLUMN_VAR.get()
            assert columns is not None
            self.columns_schema = columns
            self.export_schema(columns)
            MON_COLUMN_VAR.set(None)

        if len(set(len(q) for q in datas.values())) != 1:
            assert ValueError("All columns should have the same length.")
        self.export_data(datas)

        MON_DATAS_VAR.set(None)

    def export_schema(self, columns_schema: dict[str, dict[str, str]]) -> None:
        """
        Export schema of the data. This method should be called after all data is logged.
        """
        raise NotImplementedError()

    def export_data(self, datas: dict[str, collections.deque[DT]]) -> None:
        """
        Export data. This method should be called after all data is logged.
        """
        raise NotImplementedError()

    def log(
        self,
        data: DT,
        name: str,
        role: str,
        data_type: str,
    ) -> None:
        """
        log a data with column name, role and type to the current record
        """
        if name in self.PRESERVED_COLUMNS:
            logger.warning(
                "Column name %s is reserved, will be renamed to %s", name, name + "_"
            )
            name = name + "_"

        if role not in BENTOML_MONITOR_ROLES:
            logger.warning(
                f"Role {role} is not officially supported, but will be logged anyway."
            )
        if data_type not in BENTOML_MONITOR_TYPES:
            logger.warning(
                f"Data type {data_type} is not officially supported, but will be logged anyway."
            )

        if self.columns_schema is None:
            columns = MON_COLUMN_VAR.get()
            assert columns is not None
            if name in columns:
                logger.warning(
                    "Column name %s is duplicated, will be ignored.",
                    name,
                )
            else:
                columns[name] = {
                    "name": name,
                    "role": role,
                    "type": data_type,
                }
        datas = MON_DATAS_VAR.get()
        assert datas is not None
        datas[name].append(data)

    def log_batch(
        self,
        data_batch: t.Iterable[DT],
        name: str,
        role: str,
        data_type: str,
    ) -> None:
        """
        Log a batch of data. The data will be logged as a single column.
        """
        try:
            for data in data_batch:
                self.log(data, name, role, data_type)
        except TypeError:
            logger.warning(
                "Data batch is not iterable, will ignore the data batch. Please use log() to log a single data."
            )

    def log_table(
        self,
        data: t.Iterable[t.Iterable[DT]],
        schema: dict[str, str],
    ) -> None:
        logger.warning(
            "log_table() is not implemented yet. Will ignore the data. Please use log() or log_batch() instead."
        )
        return


class NoOpMonitor(MonitorBase[t.Any]):
    def __init__(self, name: str, **kwargs: t.Any) -> None:
        pass

    def start_record(self) -> None:
        pass

    def stop_record(self) -> None:
        pass

    def export_schema(self, columns_schema: dict[str, dict[str, str]]) -> None:
        pass

    def export_data(self, datas: dict[str, collections.deque[t.Any]]) -> None:
        pass

    def log(self, data: t.Any, name: str, role: str, data_type: str) -> None:
        pass

    def log_batch(
        self,
        data_batch: t.Iterable[t.Any],
        name: str,
        role: str,
        data_type: str,
    ) -> None:
        pass

    def log_table(
        self,
        data: t.Iterable[t.Iterable[t.Any]],
        schema: dict[str, str],
    ) -> None:
        pass



from __future__ import annotations

import importlib
import os
import shlex
import typing as t

import attrs
from fs.base import FS
from jinja2 import Environment
from jinja2 import FileSystemLoader

from bentoml._internal.bento.bento import ImageInfo
from bentoml._internal.container.generate import DEFAULT_BENTO_ENVS
from bentoml._internal.container.generate import PREHEAT_PIP_PACKAGES
from bentoml._internal.container.generate import expands_bento_path
from bentoml._internal.container.generate import resolve_package_versions
from bentoml._internal.container.generate import to_bento_field
from bentoml._internal.container.generate import to_options_field


def get_templates_variables(
    image: ImageInfo, bento_fs: FS, **bento_envs: t.Any
) -> dict[str, t.Any]:
    from bentoml._internal.configuration.containers import BentoMLContainer

    bento_envs = {**DEFAULT_BENTO_ENVS, **bento_envs}
    options = attrs.asdict(image)
    requirement_file = bento_fs.getsyspath("env/python/requirements.txt")
    if os.path.exists(requirement_file):
        python_packages = resolve_package_versions(requirement_file)
    else:
        python_packages = {}
    pip_preheat_packages = [
        python_packages[k] for k in PREHEAT_PIP_PACKAGES if k in python_packages
    ]
    return {
        **{to_options_field(k): v for k, v in options.items()},
        **{to_bento_field(k): v for k, v in bento_envs.items()},
        "__prometheus_port__": BentoMLContainer.grpc.metrics.port.get(),
        "__pip_preheat_packages__": pip_preheat_packages,
    }


def generate_dockerfile(
    image: ImageInfo, bento_fs: FS, *, frontend: str = "dockerfile", **bento_envs: t.Any
) -> str:
    templates_path = importlib.import_module(
        f"bentoml._internal.container.frontend.{frontend}.templates"
    ).__path__
    environment = Environment(
        extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols", "jinja2.ext.debug"],
        trim_blocks=True,
        lstrip_blocks=True,
        loader=FileSystemLoader(templates_path, followlinks=True),
    )
    environment.filters["bash_quote"] = shlex.quote
    environment.globals["expands_bento_path"] = expands_bento_path
    template = environment.get_template("base_v2.j2")
    return template.render(**get_templates_variables(image, bento_fs, **bento_envs))



from _bentoml_sdk.validators import ContentType
from _bentoml_sdk.validators import DataframeSchema
from _bentoml_sdk.validators import DType
from _bentoml_sdk.validators import FileSchema
from _bentoml_sdk.validators import PILImageEncoder
from _bentoml_sdk.validators import Shape
from _bentoml_sdk.validators import TensorSchema

__all__ = [
    "ContentType",
    "DataframeSchema",
    "DType",
    "FileSchema",
    "PILImageEncoder",
    "Shape",
    "TensorSchema",
]



from __future__ import annotations

import functools
import io
import typing as t
from urllib.parse import quote

from multipart.multipart import parse_options_header
from starlette.datastructures import UploadFile
from starlette.requests import Request
from starlette.responses import Response

from ...exceptions import BadInput
from ...exceptions import InternalServerError
from ...exceptions import InvalidArgument
from ...exceptions import MissingDependencyException
from ...grpc.utils import import_generated_stubs
from ..service.openapi import SUCCESS_DESCRIPTION
from ..service.openapi.specification import MediaType
from ..service.openapi.specification import Schema
from ..types import LazyType
from ..utils.filesystem import resolve_user_filepath
from ..utils.http import set_cookies
from ..utils.lazy_loader import LazyLoader
from .base import IODescriptor

PIL_EXC_MSG = "'Pillow' is required to use the Image IO descriptor. Install with 'pip install bentoml[io-image]'."

if t.TYPE_CHECKING:
    from types import UnionType

    import PIL
    import PIL.Image

    from ...grpc.v1 import service_pb2 as pb
    from ...grpc.v1alpha1 import service_pb2 as pb_v1alpha1
    from .. import external_typing as ext
    from ..context import ServiceContext as Context
    from .base import OpenAPIResponse

    _Mode = t.Literal[
        "1", "CMYK", "F", "HSV", "I", "L", "LAB", "P", "RGB", "RGBA", "RGBX", "YCbCr"
    ]
else:
    # NOTE: pillow-simd only benefits users who want to do preprocessing
    # TODO: add options for users to choose between simd and native mode
    PIL = LazyLoader("PIL", globals(), "PIL", exc_msg=PIL_EXC_MSG)
    PIL.Image = LazyLoader("PIL.Image", globals(), "PIL.Image", exc_msg=PIL_EXC_MSG)

    pb, _ = import_generated_stubs("v1")
    pb_v1alpha1, _ = import_generated_stubs("v1alpha1")

# NOTES: we will keep type in quotation to avoid backward compatibility
#  with numpy < 1.20, since we will use the latest stubs from the main branch of numpy.
#  that enable a new way to type hint an ndarray.
ImageType = t.Union["PIL.Image.Image", "ext.NpNDArray"]

DEFAULT_PIL_MODE = "RGB"


PIL_WRITE_ONLY_FORMATS = {"PALM", "PDF"}
READABLE_MIMES: set[str] = None  # type: ignore (lazy constant)
MIME_EXT_MAPPING: dict[str, str] = None  # type: ignore (lazy constant)


@functools.lru_cache(maxsize=1)
def initialize_pillow():
    global MIME_EXT_MAPPING  # pylint: disable=global-statement
    global READABLE_MIMES  # pylint: disable=global-statement

    try:
        import PIL.Image
    except ImportError:
        raise InternalServerError(PIL_EXC_MSG)

    PIL.Image.init()
    MIME_EXT_MAPPING = {v: k for k, v in PIL.Image.MIME.items()}  # type: ignore (lazy constant)
    READABLE_MIMES = {
        k for k, v in MIME_EXT_MAPPING.items() if v not in PIL_WRITE_ONLY_FORMATS
    }  # type: ignore (lazy constant)


class Image(
    IODescriptor[ImageType], descriptor_id="bentoml.io.Image", proto_fields=("file",)
):
    """
    :obj:`Image` defines API specification for the inputs/outputs of a Service, where either
    inputs will be converted to or outputs will be converted from images as specified
    in your API function signature.

    A sample object detection service:

    .. code-block:: python
       :caption: `service.py`

       from __future__ import annotations

       from typing import TYPE_CHECKING
       from typing import Any

       import bentoml
       from bentoml.io import Image
       from bentoml.io import NumpyNdarray

       if TYPE_CHECKING:
           from PIL.Image import Image
           from numpy.typing import NDArray

       runner = bentoml.tensorflow.get('image-classification:latest').to_runner()

       svc = bentoml.Service("vit-object-detection", runners=[runner])

       @svc.api(input=Image(), output=NumpyNdarray(dtype="float32"))
       async def predict_image(f: Image) -> NDArray[Any]:
           assert isinstance(f, Image)
           arr = np.array(f) / 255.0
           assert arr.shape == (28, 28)

           # We are using greyscale image and our PyTorch model expect one
           # extra channel dimension
           arr = np.expand_dims(arr, (0, 3)).astype("float32")  # reshape to [1, 28, 28, 1]
           return await runner.async_run(arr)

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./service.py:svc --reload

    Users can then send requests to the newly started services with any client:

    .. tab-set::

        .. tab-item:: Bash

            .. code-block:: bash

               # we will run on our input image test.png
               # image can get from http://images.cocodataset.org/val2017/000000039769.jpg
               % curl -H "Content-Type: multipart/form-data" \\
                      -F 'fileobj=@test.jpg;type=image/jpeg' \\
                      http://0.0.0.0:3000/predict_image

               # [{"score":0.8610631227493286,"label":"Egyptian cat"},
               # {"score":0.08770329505205154,"label":"tabby, tabby cat"},
               # {"score":0.03540956228971481,"label":"tiger cat"},
               # {"score":0.004140055272728205,"label":"lynx, catamount"},
               # {"score":0.0009498853469267488,"label":"Siamese cat, Siamese"}]%

        .. tab-item:: Python

           .. code-block:: python
              :caption: `request.py`

              import requests

              requests.post(
                  "http://0.0.0.0:3000/predict_image",
                  files = {"upload_file": open('test.jpg', 'rb')},
                  headers = {"content-type": "multipart/form-data"}
              ).text

    Args:
        pilmode: Color mode for PIL. Default to ``RGB``.
        mime_type: The MIME type of the file type that this descriptor should return. Only relevant when used as an output descriptor.
        allowed_mime_types: A list of MIME types to restrict input to.

    Returns:
        :obj:`Image`: IO Descriptor that either a :code:`PIL.Image.Image` or a :code:`np.ndarray` representing an image.
    """

    def __init__(
        self,
        pilmode: _Mode | None = DEFAULT_PIL_MODE,
        mime_type: str = "image/jpeg",
        *,
        allowed_mime_types: t.Iterable[str] | None = None,
    ):
        initialize_pillow()

        if pilmode is not None and pilmode not in PIL.Image.MODES:  # pragma: no cover
            raise InvalidArgument(
                f"Invalid Image pilmode '{pilmode}'. Supported PIL modes are {', '.join(PIL.Image.MODES)}."
            ) from None

        self._mime_type = mime_type.lower()
        self._allowed_mimes: set[str] = (
            READABLE_MIMES
            if allowed_mime_types is None
            else {mtype.lower() for mtype in allowed_mime_types}
        )
        self._allow_all_images = allowed_mime_types is None

        if self._mime_type not in MIME_EXT_MAPPING:  # pragma: no cover
            raise InvalidArgument(
                f"Invalid Image mime_type '{mime_type}'; supported mime types are {', '.join(PIL.Image.MIME.values())} "
            )

        self._pilmode: _Mode | None = pilmode
        self._format: str = MIME_EXT_MAPPING[self._mime_type]

    def _from_sample(self, sample: ImageType | str) -> ImageType:
        """
        Create a :class:`~bentoml._internal.io_descriptors.image.Image` IO Descriptor from given inputs.

        Args:
            sample: Given File-like object, or a path to a file.
            pilmode: Optional color mode for PIL. Default to ``RGB``.
            mime_type: The MIME type of the file type that this descriptor should return.
                       If not specified, then ``from_sample`` will try to infer the MIME type
                       from file extension.
            allowed_mime_types: An optional list of MIME types to restrict input to.

        Returns:
            :class:`~bentoml._internal.io_descriptors.image.Image`: IODescriptor from given users inputs.

        Example:

        .. code-block:: python
           :caption: `service.py`

           from __future__ import annotations

           import bentoml
           from typing import Any
           from bentoml.io import Image
           import numpy as np

           input_spec = Image.from_sample("/path/to/image.jpg")
           @svc.api(input=input_spec, output=Image())
           async def predict(input: t.IO[t.Any]) -> t.IO[t.Any]:
               return await runner.async_run(input)

        Raises:
            :class:`InvalidArgument`: If the given sample is not a valid image type.
            :class:`MissingDependencyException`: If ``filetype`` is not installed.
            :class:`BadInput`: Given sample from file can't be parsed with Pillow.
        """
        try:
            from filetype.match import image_match
        except ModuleNotFoundError:
            raise MissingDependencyException(
                "'filetype' is required to use 'from_sample'. Install it with 'pip install bentoml[io-image]'."
            )

        img_type = image_match(sample)
        if img_type is None:
            raise InvalidArgument(f"{sample} is not a valid image file type.")

        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(sample):
            sample = PIL.Image.fromarray(sample)
        elif isinstance(sample, str):
            p = resolve_user_filepath(sample, ctx=None)
            try:
                with open(p, "rb") as f:
                    sample = PIL.Image.open(f)
            except PIL.UnidentifiedImageError as err:
                raise BadInput(f"Failed to parse sample image file: {err}") from None
        self._mime_type = img_type.mime
        return sample

    def to_spec(self) -> dict[str, t.Any]:
        return {
            "id": self.descriptor_id,
            "args": {
                "pilmode": self._pilmode,
                "mime_type": self._mime_type,
                "allowed_mime_types": list(self._allowed_mimes),
            },
        }

    @classmethod
    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
        if "args" not in spec:
            raise InvalidArgument(f"Missing args key in Image spec: {spec}")

        return cls(**spec["args"])

    def input_type(self) -> UnionType:
        return ImageType

    def openapi_schema(self) -> Schema:
        return Schema(type="string", format="binary")

    def openapi_components(self) -> dict[str, t.Any] | None:
        pass

    def openapi_example(self):
        pass

    def openapi_request_body(self) -> dict[str, t.Any]:
        return {
            "content": {
                mtype: MediaType(schema=self.openapi_schema())
                for mtype in self._allowed_mimes
            },
            "required": True,
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    def openapi_responses(self) -> OpenAPIResponse:
        return {
            "description": SUCCESS_DESCRIPTION,
            "content": {self._mime_type: MediaType(schema=self.openapi_schema())},
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    async def from_http_request(self, request: Request) -> ImageType:
        content_type, _ = parse_options_header(request.headers["content-type"])
        mime_type = content_type.decode().lower()

        bytes_: bytes | str | None = None

        if mime_type == "multipart/form-data":
            form = await request.form()

            found_mimes: list[str] = []

            for val in form.values():
                val_content_type = val.content_type  # type: ignore (bad starlette types)
                if isinstance(val, UploadFile):
                    found_mimes.append(val_content_type)

                if self._allowed_mimes is None:
                    if (
                        val_content_type in MIME_EXT_MAPPING
                        or val_content_type.startswith("image/")
                    ):
                        bytes_ = await val.read()
                        break
                elif val_content_type in self._allowed_mimes:
                    bytes_ = await val.read()
                    break
            else:
                if len(found_mimes) == 0:
                    raise BadInput("no image file found in multipart form")
                else:
                    if self._allowed_mimes is None:
                        raise BadInput(
                            f"no multipart image file (supported images are: {', '.join(MIME_EXT_MAPPING.keys())}, or 'image/*'), got files with content types {', '.join(found_mimes)}"
                        )
                    else:
                        raise BadInput(
                            f"no multipart image file (allowed mime types are: {', '.join(self._allowed_mimes)}), got files with content types {', '.join(found_mimes)}"
                        )

        elif self._allowed_mimes is None:
            if mime_type in MIME_EXT_MAPPING or mime_type.startswith("image/"):
                bytes_ = await request.body()
        elif mime_type in self._allowed_mimes:
            bytes_ = await request.body()
        else:
            if self._allowed_mimes is None:
                raise BadInput(
                    f"unsupported mime type {mime_type}; supported mime types are: {', '.join(MIME_EXT_MAPPING.keys())}, or 'image/*'"
                )
            else:
                raise BadInput(
                    f"mime type {mime_type} is not allowed, allowed mime types are: {', '.join(self._allowed_mimes)}"
                )

        assert bytes_ is not None

        if isinstance(bytes_, str):
            bytes_ = bytes(bytes_, "UTF-8")

        try:
            return PIL.Image.open(io.BytesIO(bytes_))
        except PIL.UnidentifiedImageError as err:
            raise BadInput(f"Failed to parse uploaded image file: {err}") from None

    async def to_http_response(
        self, obj: ImageType, ctx: Context | None = None
    ) -> Response:
        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(obj):
            image = PIL.Image.fromarray(obj, mode=self._pilmode)
        elif LazyType["PIL.Image.Image"]("PIL.Image.Image").isinstance(obj):
            image = obj
        else:
            raise BadInput(
                f"Unsupported Image type received: '{type(obj)}', the Image IO descriptor only supports 'np.ndarray' and 'PIL.Image'."
            ) from None
        filename = f"output.{self._format.lower()}"

        ret = io.BytesIO()
        image.save(ret, format=self._format)

        # rfc2183
        content_disposition_filename = quote(filename)
        if content_disposition_filename != filename:
            content_disposition = "attachment; filename*=utf-8''{}".format(
                content_disposition_filename
            )
        else:
            content_disposition = f'attachment; filename="{filename}"'

        if ctx is not None:
            if "content-disposition" not in ctx.response.headers:
                ctx.response.headers["content-disposition"] = content_disposition
            res = Response(
                ret.getvalue(),
                media_type=self._mime_type,
                headers=ctx.response.headers,  # type: ignore (bad starlette types)
                status_code=ctx.response.status_code,
            )
            set_cookies(res, ctx.response.cookies)
            return res
        else:
            return Response(
                ret.getvalue(),
                media_type=self._mime_type,
                headers={"content-disposition": content_disposition},
            )

    async def from_proto(self, field: pb.File | pb_v1alpha1.File | bytes) -> ImageType:
        if isinstance(field, bytes):
            content = field
        elif isinstance(field, pb_v1alpha1.File):
            from .file import filetype_pb_to_mimetype_map

            mapping = filetype_pb_to_mimetype_map()
            if field.kind:
                try:
                    mime_type = mapping[field.kind]
                    if mime_type != self._mime_type:
                        raise BadInput(
                            f"Inferred mime_type from 'kind' is '{mime_type}', while '{self!r}' is expecting '{self._mime_type}'",
                        )
                except KeyError:
                    raise BadInput(
                        f"{field.kind} is not a valid File kind. Accepted file kind: {[names for names, _ in pb_v1alpha1.File.FileType.items()]}",
                    ) from None
            if not field.content:
                raise BadInput("Content is empty!") from None
            return PIL.Image.open(io.BytesIO(field.content))
        else:
            assert isinstance(field, pb.File)
            if field.kind and field.kind != self._mime_type:
                raise BadInput(
                    f"MIME type from 'kind' is '{field.kind}', while '{self!r}' is expecting '{self._mime_type}'",
                )
            content = field.content
            if not content:
                raise BadInput("Content is empty!") from None

        return PIL.Image.open(io.BytesIO(content))

    async def to_proto_v1alpha1(self, obj: ImageType) -> pb_v1alpha1.File:
        from .file import mimetype_to_filetype_pb_map

        try:
            kind = mimetype_to_filetype_pb_map()[self._mime_type]
        except KeyError:
            raise BadInput(
                f"{self._mime_type} doesn't have a corresponding File 'kind'"
            ) from None

        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(obj):
            image = PIL.Image.fromarray(obj, mode=self._pilmode)
        elif LazyType["PIL.Image.Image"]("PIL.Image.Image").isinstance(obj):
            image = obj
        else:
            raise BadInput(
                f"Unsupported Image type received: '{type(obj)}', the Image IO descriptor only supports 'np.ndarray' and 'PIL.Image'.",
            ) from None
        ret = io.BytesIO()
        image.save(ret, format=self._format)

        return pb_v1alpha1.File(kind=kind, content=ret.getvalue())

    async def to_proto(self, obj: ImageType) -> pb.File:
        if LazyType["ext.NpNDArray"]("numpy.ndarray").isinstance(obj):
            image = PIL.Image.fromarray(obj, mode=self._pilmode)
        elif LazyType["PIL.Image.Image"]("PIL.Image.Image").isinstance(obj):
            image = obj
        else:
            raise BadInput(
                f"Unsupported Image type received: '{type(obj)}', the Image IO descriptor only supports 'np.ndarray' and 'PIL.Image'.",
            ) from None
        ret = io.BytesIO()
        image.save(ret, format=self._format)

        return pb.File(kind=self._mime_type, content=ret.getvalue())



from __future__ import annotations

import typing as t
from datetime import datetime

import attr

from bentoml._internal.cloud.schemas.modelschemas import BentoManifestSchema
from bentoml._internal.cloud.schemas.modelschemas import DeploymentTargetCanaryRule
from bentoml._internal.cloud.schemas.modelschemas import DeploymentTargetConfig
from bentoml._internal.cloud.schemas.modelschemas import LabelItemSchema
from bentoml._internal.cloud.schemas.modelschemas import ModelManifestSchema
from bentoml._internal.cloud.schemas.modelschemas import ResourceInstanceConfigSchema
from bentoml._internal.cloud.schemas.modelschemas import TransmissionStrategy


@attr.define
class BaseSchema:
    uid: str
    created_at: datetime
    updated_at: t.Optional[datetime]
    deleted_at: t.Optional[datetime]


@attr.define
class BaseListSchema:
    start: int
    count: int
    total: int


@attr.define
class ResourceSchema(BaseSchema):
    name: str
    resource_type: str
    labels: t.List[LabelItemSchema]


@attr.define
class UserSchema:
    name: str
    email: str
    first_name: str
    last_name: str

    def get_name(self) -> str:
        if not self.first_name and not self.last_name:
            return self.name
        return f"{self.first_name} {self.last_name}".strip()


@attr.define
class OrganizationSchema(ResourceSchema):
    description: str


@attr.define
class OrganizationListSchema(BaseListSchema):
    items: t.List[OrganizationSchema]


@attr.define
class ClusterSchema(ResourceSchema):
    description: str
    organization_name: str
    creator: UserSchema
    is_first: t.Optional[bool] = None


@attr.define
class ClusterConfigSchema:
    default_deployment_kube_namespace: str


@attr.define
class ClusterListSchema(BaseListSchema):
    items: t.List[ClusterSchema]


@attr.define
class CreateBentoRepositorySchema:
    name: str
    description: str


@attr.define
class BentoSchema(ResourceSchema):
    description: str
    version: str
    image_build_status: str
    upload_status: str
    upload_finished_reason: str
    presigned_upload_url: str
    presigned_download_url: str
    manifest: t.Optional[BentoManifestSchema] = attr.field(default=None)
    transmission_strategy: t.Optional[TransmissionStrategy] = attr.field(default=None)
    upload_id: t.Optional[str] = attr.field(default=None)

    upload_started_at: t.Optional[datetime] = attr.field(default=None)
    upload_finished_at: t.Optional[datetime] = attr.field(default=None)
    build_at: datetime = attr.field(factory=datetime.now)


@attr.define
class BentoRepositorySchema(ResourceSchema):
    description: str
    latest_bento: t.Optional[BentoSchema]


@attr.define
class BentoWithRepositorySchema(BentoSchema):
    repository: BentoRepositorySchema = attr.field(default=None)


@attr.define
class BentoWithRepositoryListSchema(BaseListSchema):
    items: t.List[BentoWithRepositorySchema] = attr.field(factory=list)


@attr.define
class CreateBentoSchema:
    description: str
    version: str
    manifest: t.Optional[BentoManifestSchema] = attr.field(default=None)
    build_at: datetime = attr.field(factory=datetime.now)
    labels: t.List[LabelItemSchema] = attr.field(factory=list)


@attr.define
class UpdateBentoSchema:
    description: t.Optional[str] = attr.field(default=None)
    manifest: t.Optional[BentoManifestSchema] = attr.field(default=None)
    labels: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)


@attr.define
class BentoFullSchema(BentoWithRepositorySchema):
    models: t.List[ModelWithRepositorySchema] = attr.field(factory=list)


@attr.define
class PreSignMultipartUploadUrlSchema:
    upload_id: str
    part_number: int


@attr.define
class CompletePartSchema:
    part_number: int
    etag: str


@attr.define
class CompleteMultipartUploadSchema:
    parts: t.List[CompletePartSchema]
    upload_id: str


@attr.define
class FinishUploadSchema:
    status: t.Optional[str]
    reason: t.Optional[str]


@attr.define
class CreateModelRepositorySchema:
    name: str
    description: str


@attr.define
class ModelSchema(ResourceSchema):
    description: str
    version: str
    image_build_status: str
    upload_status: str
    upload_finished_reason: str
    presigned_upload_url: str
    presigned_download_url: str
    manifest: ModelManifestSchema

    transmission_strategy: t.Optional[TransmissionStrategy] = attr.field(default=None)
    upload_id: t.Optional[str] = attr.field(default=None)

    upload_started_at: t.Optional[datetime] = attr.field(default=None)
    upload_finished_at: t.Optional[datetime] = attr.field(default=None)
    build_at: datetime = attr.field(factory=datetime.now)


@attr.define
class ModelRepositorySchema(ResourceSchema):
    description: str
    latest_model: t.Optional[ModelSchema]


@attr.define
class ModelWithRepositorySchema(ModelSchema):
    repository: ModelRepositorySchema = attr.field(default=None)


@attr.define
class ModelWithRepositoryListSchema(BaseListSchema):
    items: t.List[ModelWithRepositorySchema] = attr.field(factory=list)


@attr.define
class CreateModelSchema:
    description: str
    version: str
    manifest: ModelManifestSchema
    build_at: datetime = attr.field(factory=datetime.now)
    labels: t.List[LabelItemSchema] = attr.field(factory=list)


@attr.define
class BentoRepositoryListSchema(BaseListSchema):
    items: t.List[BentoRepositorySchema]


@attr.define
class BentoListSchema(BaseListSchema):
    items: t.List[BentoSchema]


@attr.define
class CreateDeploymentTargetSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    bento_repository: str
    bento: str
    config: DeploymentTargetConfig
    canary_rules: t.Optional[t.List[DeploymentTargetCanaryRule]] = attr.field(
        default=None
    )


@attr.define
class DeploymentSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    creator: UserSchema
    cluster: ClusterSchema
    status: str
    kube_namespace: str
    latest_revision: t.Optional[DeploymentRevisionSchema] = attr.field(
        default=None
    )  # Delete returns no latest revision


@attr.define
class DeploymentTargetSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    creator: UserSchema
    bento: BentoFullSchema
    config: DeploymentTargetConfig
    canary_rules: t.Optional[t.List[DeploymentTargetCanaryRule]] = attr.field(
        default=None
    )


@attr.define
class DeploymentRevisionSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    creator: UserSchema
    status: str
    targets: t.List[DeploymentTargetSchema]


@attr.define
class ResourceInstanceSchema(ResourceSchema):
    display_name: str
    description: str
    config: ResourceInstanceConfigSchema


@attr.define(kw_only=True)
class ClusterFullSchema(ClusterSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    organization: OrganizationSchema
    kube_config: str
    config: ClusterConfigSchema
    grafana_root_path: str
    resource_instances: t.List[ResourceInstanceSchema]


@attr.define
class DeploymentListSchema(BaseListSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    items: t.List[DeploymentSchema]


@attr.define
class UpdateDeploymentSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    targets: t.List[CreateDeploymentTargetSchema]
    labels: t.Optional[t.List[LabelItemSchema]] = attr.field(default=None)
    description: t.Optional[str] = attr.field(default=None)
    do_not_deploy: t.Optional[bool] = attr.field(default=None)


@attr.define(kw_only=True)
class CreateDeploymentSchema(UpdateDeploymentSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    name: str
    kube_namespace: str


@attr.define(kw_only=True)
class DeploymentFullSchema(DeploymentSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    urls: list[str]


@attr.define
class SecretItem:
    key: str
    sub_path: t.Optional[str] = attr.field(default=None)
    value: t.Optional[str] = attr.field(default=None)


@attr.define
class SecretContentSchema:
    type: str
    items: t.List[SecretItem]
    path: t.Optional[str] = attr.field(default=None)


@attr.define
class SecretSchema(ResourceSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    description: str
    creator: UserSchema
    content: SecretContentSchema
    cluster: ClusterSchema


@attr.define
class SecretListSchema(BaseListSchema):
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    items: t.List[SecretSchema]


@attr.define
class CreateSecretSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    name: str
    content: SecretContentSchema
    description: t.Optional[str] = attr.field(default=None)


@attr.define
class UpdateSecretSchema:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    content: SecretContentSchema
    description: t.Optional[str] = attr.field(default=None)



from __future__ import annotations

from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from typing import Optional
from urllib.parse import parse_qs


class AuthRedirectHandler(BaseHTTPRequestHandler):
    """
    HTTPRequest Handler that is intended to be used as oauth2 callback page
    """

    def log_message(self, format, *args):
        """
        Disables logging for BaseHTTPRequestHandler
        """
        # silence the log messages
        pass

    def end_headers(self):
        """Add necessary headers for CORS"""
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Methods", "POST, OPTIONS")
        self.send_header("Access-Control-Allow-Headers", "Content-Type")
        super().end_headers()

    def do_OPTIONS(self):
        """Handle preflight CORS requests"""
        self.send_response(200)
        self.end_headers()

    def do_POST(self) -> None:
        """
        Provides callback page for the oauth2 redirect using POST request
        """
        try:
            content_length = int(self.headers["Content-Length"])
            post_data = self.rfile.read(content_length).decode("utf-8")
            params = parse_qs(post_data)

            has_error = (
                "code" not in params
                or len(params["code"]) != 1
                or params["code"][0].strip() == ""
            )

            if has_error:
                self.send_response(400)
                self.end_headers()
            else:
                self.send_response(200)
                self.end_headers()
                self.server._code = params["code"][0]
        except Exception:
            self.send_response(500)
            self.end_headers()


class AuthCallbackHttpServer(HTTPServer):
    """
    Simplistic HTTP Server to provide local callback URL for token provider
    """

    def __init__(self, port: int):
        super().__init__(("", port), AuthRedirectHandler)

        self._code: str | None = None

    def get_code(self) -> str | None:
        """
        This method should only be called after the request was done and might be None when no token is given.

        :return: Authorization code or None if the request was not performed yet
        """
        return self._code

    @property
    def callback_url(self) -> str:
        """
        Callback URL for the HTTP-Server
        """
        return f"http://localhost:{self.server_port}"

    def wait_indefinitely_for_code(self) -> Optional[str]:
        """
        Wait indefinitely for ther server to callback from token provider.
        """
        while self._code is None:
            self.handle_request()
        return self.get_code()



import typing as t

from tensorflow import Tensor
from tensorflow.python.client.session import Session
from tensorflow.python.eager.def_function import Function
from tensorflow.python.eager.function_spec import FunctionSpec
from tensorflow.python.framework.composite_tensor import CompositeTensor
from tensorflow.python.framework.indexed_slices import IndexedSlices
from tensorflow.python.framework.sparse_tensor import SparseTensorSpec
from tensorflow.python.framework.tensor_spec import TensorSpec
from tensorflow.python.framework.type_spec import TypeSpec
from tensorflow.python.keras.engine.sequential import Sequential
from tensorflow.python.keras.engine.training import Model
from tensorflow.python.module.module import Module
from tensorflow.python.ops.ragged.ragged_tensor import RaggedTensorSpec
from tensorflow.python.ops.tensor_array_ops import TensorArray
from tensorflow.python.ops.tensor_array_ops import TensorArraySpec
from tensorflow.python.saved_model.function_deserialization import RestoredFunction
from tensorflow.python.saved_model.save_options import SaveOptions
from tensorflow.python.trackable.base import Trackable
from tensorflow.python.training.tracking.tracking import AutoTrackable

# NOTE: FunctionSpec moved from eager.function to eager.function_spec
# and then to eager.polymorphic_function
try:
    from tensorflow.python.eager.function import FunctionSpec
except ImportError:
    try:
        from tensorflow.python.eager.function_spec import FunctionSpec
    except ImportError:
        from tensorflow.python.eager.polymorphic_function import FunctionSpec

try:
    from tensorflow.python.framework.ops import _EagerTensorBase as EagerTensor
    from tensorflow.python.types.core import ConcreteFunction
    from tensorflow.python.types.core import GenericFunction
except ImportError:
    from tensorflow.python.eager.function import ConcreteFunction

    class GenericFunction(t.Protocol):
        def __call__(self, *args: t.Any, **kwargs: t.Any): ...

    # NOTE: future proof when tensorflow decided to remove EagerTensor
    # and enable eager execution on Tensor by default. This class only serves
    # as type fallback.
    class EagerTensor(Tensor): ...


class SignatureMap(t.Mapping[str, t.Any], Trackable):
    """A collection of SavedModel signatures."""

    _signatures: t.Dict[str, t.Union[ConcreteFunction, RestoredFunction, Function]]

    def __init__(self) -> None: ...

    def __getitem__(
        self, key: str
    ) -> t.Union[ConcreteFunction, RestoredFunction, Function]: ...

    def __iter__(self) -> t.Iterator[str]: ...

    def __len__(self) -> int: ...

    def __repr__(self) -> str: ...


# This denotes all decorated function with `tf.function`
DecoratedFunction = t.Union[RestoredFunction, ConcreteFunction, GenericFunction]

# This denotes all possible tensor type that can be casted with `tf.cast`
CastableTensorType = t.Union[Tensor, EagerTensor, CompositeTensor, IndexedSlices]
# This encapsulates scenarios where input can be accepted as TensorArray
TensorLike = t.Union[CastableTensorType, TensorArray]
# This defines all possible TensorSpec from tensorflow API
UnionTensorSpec = t.Union[
    TensorSpec,
    RaggedTensorSpec,
    SparseTensorSpec,
    TensorArraySpec,
]

# TODO(aarnphm): Specify types instead of t.Any
TensorSignature = t.Tuple[TensorSpec, bool, t.Optional[t.Any]]
InputSignature = t.Tuple[TensorSignature, t.Dict[str, TypeSpec]]

# This denotes all Keras Model API
KerasModel = t.Union[Model, Sequential]

__all__ = [
    "EagerTensor",
    "CastableTensorType",
    "TensorLike",
    "InputSignature",
    "TensorSignature",
    "UnionTensorSpec",
    "Trackable",
    "AutoTrackable",
    "ConcreteFunction",
    "RestoredFunction",
    "FunctionSpec",
    "SignatureMap",
    "Module",
    "TypeSpec",
    "SaveOptions",
    "Session",
    "KerasModel",
]
